# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information

# scapy.contrib.description = Link Layer Discovery Protocol (LLDP)
# scapy.contrib.status = loads

"""
    LLDP - Link Layer Discovery Protocol
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :author:    Thomas Tannhaeuser, hecke@naberius.de

    :description:

        This module provides Scapy layers for the LLDP protocol.

        normative references:
            - IEEE 802.1AB 2016 - LLDP protocol, topology and MIB description

    :TODO:
        - | organization specific TLV e.g. ProfiNet
          | (see LLDPDUGenericOrganisationSpecific for a starting point)
        - Ignore everything after EndofLLDPDUTLV

    :NOTES:
        - you can find the layer configuration options at the end of this file
        - default configuration enforces standard conform:

          * | frame structure
            | (ChassisIDTLV/PortIDTLV/TimeToLiveTLV/...)
          * multiplicity of TLVs (if given by the standard)
          * min sizes of strings used by the TLVs

        - conf.contribs['LLDP'].strict_mode_disable() -> disable strict mode

"""
from scapy.config import conf
from scapy.error import Scapy_Exception
from scapy.layers.l2 import Ether, Dot1Q
from scapy.fields import MACField, IPField, IP6Field, BitField, \
    StrLenField, ByteEnumField, BitEnumField, \
    EnumField, ThreeBytesField, BitFieldLenField, \
    ShortField, XStrLenField, ByteField, ConditionalField, \
    MultipleTypeField, FlagsField, ShortEnumField, ScalingField, \
    BitScalingField
from scapy.packet import Packet, bind_layers
from scapy.data import ETHER_TYPES
from scapy.compat import orb, bytes_int

LLDP_NEAREST_BRIDGE_MAC = '01:80:c2:00:00:0e'
LLDP_NEAREST_NON_TPMR_BRIDGE_MAC = '01:80:c2:00:00:03'
LLDP_NEAREST_CUSTOMER_BRIDGE_MAC = '01:80:c2:00:00:00'

LLDP_ETHER_TYPE = 0x88cc
ETHER_TYPES[LLDP_ETHER_TYPE] = 'LLDP'


class LLDPInvalidFieldValue(Scapy_Exception):
    """
    field value is out of allowed range
    """
    pass


class LLDPInvalidFrameStructure(Scapy_Exception):
    """
    basic frame structure not standard conform
    (missing TLV, invalid order or multiplicity)
    """
    pass


class LLDPMissingLowerLayer(Scapy_Exception):
    """
    first layer below first LLDPDU must be Ethernet or Dot1q
    """
    pass


class LLDPInvalidTLVCount(Scapy_Exception):
    """
    invalid number of entries for a specific TLV type
    """
    pass


class LLDPInvalidLengthField(Scapy_Exception):
    """
    invalid value of length field
    """
    pass


class LLDPDU(Packet):
    """
    base class for all LLDP data units
    """
    TYPES = {
        0x00: 'end of LLDPDU',
        0x01: 'chassis id',
        0x02: 'port id',
        0x03: 'time to live',
        0x04: 'port description',
        0x05: 'system name',
        0x06: 'system description',
        0x07: 'system capabilities',
        0x08: 'management address',
        127: 'organisation specific TLV'
    }

    IANA_ADDRESS_FAMILY_NUMBERS = {
        0x00: 'other',
        0x01: 'IPv4',
        0x02: 'IPv6',
        0x03: 'NSAP',
        0x04: 'HDLC',
        0x05: 'BBN',
        0x06: '802',
        0x07: 'E.163',
        0x08: 'E.164',
        0x09: 'F.69',
        0x0a: 'X.121',
        0x0b: 'IPX',
        0x0c: 'Appletalk',
        0x0d: 'Decnet IV',
        0x0e: 'Banyan Vines',
        0x0f: 'E.164 with NSAP',
        0x10: 'DNS',
        0x11: 'Distinguished Name',
        0x12: 'AS Number',
        0x13: 'XTP over IPv4',
        0x14: 'XTP over IPv6',
        0x15: 'XTP native mode XTP',
        0x16: 'Fiber Channel World-Wide Port Name',
        0x17: 'Fiber Channel World-Wide Node Name',
        0x18: 'GWID',
        0x19: 'AFI for L2VPN',
        0x1a: 'MPLS-TP Section Endpoint ID',
        0x1b: 'MPLS-TP LSP Endpoint ID',
        0x1c: 'MPLS-TP Pseudowire Endpoint ID',
        0x1d: 'MT IP Multi-Topology IPv4',
        0x1e: 'MT IP Multi-Topology IPv6'
    }

    DOT1Q_HEADER_LEN = 4
    ETHER_HEADER_LEN = 14
    ETHER_FSC_LEN = 4
    ETHER_FRAME_MIN_LEN = 64

    LAYER_STACK = []
    LAYER_MULTIPLICITIES = {}

    def guess_payload_class(self, payload):
        # type is a 7-bit bitfield spanning bits 1..7 -> div 2
        try:
            lldpdu_tlv_type = orb(payload[0]) // 2
            class_type = LLDPDU_CLASS_TYPES.get(lldpdu_tlv_type, conf.raw_layer)
            if isinstance(class_type, list):
                for cls in class_type:
                    if cls._match_organization_specific(payload):
                        return cls
            else:
                return class_type
        except IndexError:
            return conf.raw_layer

    @staticmethod
    def _dot1q_headers_size(layer):
        """
        calculate size of lower dot1q layers (if present)
        :param layer: the layer to start at
        :return: size of vlan headers, layer below lowest vlan header
        """

        vlan_headers_size = 0
        under_layer = layer

        while under_layer and isinstance(under_layer, Dot1Q):
            vlan_headers_size += LLDPDU.DOT1Q_HEADER_LEN
            under_layer = under_layer.underlayer

        return vlan_headers_size, under_layer

    def post_build(self, pkt, pay):

        under_layer = self.underlayer

        if under_layer is None:
            if conf.contribs['LLDP'].strict_mode():
                raise LLDPMissingLowerLayer('No lower layer (Ethernet '
                                            'or Dot1Q) provided.')
            else:
                return pkt + pay

        if isinstance(under_layer, LLDPDU):
            return pkt + pay

        frame_size, under_layer = LLDPDU._dot1q_headers_size(under_layer)

        if not under_layer or not isinstance(under_layer, Ether):
            if conf.contribs['LLDP'].strict_mode():
                raise LLDPMissingLowerLayer('No Ethernet layer provided.')
            else:
                return pkt + pay

        frame_size += LLDPDU.ETHER_HEADER_LEN
        frame_size += len(pkt) + len(pay) + LLDPDU.ETHER_FSC_LEN
        if frame_size < LLDPDU.ETHER_FRAME_MIN_LEN:
            return pkt + pay + b'\x00' * (LLDPDU.ETHER_FRAME_MIN_LEN - frame_size)  # noqa: E501
        return pkt + pay

    @staticmethod
    def _frame_structure_check(structure_description):
        """
        check if the structure of the frame is conform to the basic
        frame structure defined by the standard
        :param structure_description: string-list reflecting LLDP-msg structure
        """

        standard_frame_structure = [LLDPDUChassisID.__name__,
                                    LLDPDUPortID.__name__,
                                    LLDPDUTimeToLive.__name__,
                                    '<...>']

        if len(structure_description) < 3:
            raise LLDPInvalidFrameStructure(
                'Invalid frame structure.\ngot: {}\nexpected: '
                '{}'.format(' '.join(structure_description),
                            ' '.join(standard_frame_structure)))

        for idx, layer_name in enumerate(standard_frame_structure):

            if layer_name == '<...>':
                break
            if layer_name != structure_description[idx]:
                raise LLDPInvalidFrameStructure(
                    'Invalid frame structure.\ngot: {}\nexpected: '
                    '{}'.format(' '.join(structure_description),
                                ' '.join(standard_frame_structure)))

    @staticmethod
    def _tlv_multiplicities_check(tlv_type_count):
        """
        check if multiplicity of present TLVs conforms to the standard
        :param tlv_type_count: dict containing counte-per-TLV
        """

        # * : 0..n, 1 : one and only one.
        standard_multiplicities = {
            LLDPDUEndOfLLDPDU.__name__: '*',
            LLDPDUChassisID.__name__: 1,
            LLDPDUPortID.__name__: 1,
            LLDPDUTimeToLive.__name__: 1,
            LLDPDUPortDescription: '*',
            LLDPDUSystemName: '*',
            LLDPDUSystemDescription: '*',
            LLDPDUSystemCapabilities: '*',
            LLDPDUManagementAddress: '*'
        }

        for tlv_type_name in standard_multiplicities:

            standard_tlv_multiplicity = \
                standard_multiplicities[tlv_type_name]
            if standard_tlv_multiplicity == '*':
                continue

            try:
                if tlv_type_count[tlv_type_name] != standard_tlv_multiplicity:
                    raise LLDPInvalidTLVCount(
                        'Invalid number of entries for TLV type '
                        '{} - expected {} entries, got '
                        '{}'.format(tlv_type_name,
                                    standard_tlv_multiplicity,
                                    tlv_type_count[tlv_type_name]))

            except KeyError:
                raise LLDPInvalidTLVCount('Missing TLV layer of type '
                                          '{}.'.format(tlv_type_name))

    def pre_dissect(self, s):

        if conf.contribs['LLDP'].strict_mode():
            if self.__class__.__name__ == 'LLDPDU':
                LLDPDU.LAYER_STACK = []
                LLDPDU.LAYER_MULTIPLICITIES = {}
            else:
                LLDPDU.LAYER_STACK.append(self.__class__.__name__)
                try:
                    LLDPDU.LAYER_MULTIPLICITIES[self.__class__.__name__] += 1
                except KeyError:
                    LLDPDU.LAYER_MULTIPLICITIES[self.__class__.__name__] = 1

        return s

    def dissection_done(self, pkt):

        if self.__class__.__name__ == 'LLDPDU' and \
                conf.contribs['LLDP'].strict_mode():
            LLDPDU._frame_structure_check(LLDPDU.LAYER_STACK)
            LLDPDU._tlv_multiplicities_check(LLDPDU.LAYER_MULTIPLICITIES)

        super(LLDPDU, self).dissection_done(pkt)

    def _check(self):
        """Overwritten by LLDPU objects"""
        pass

    def post_dissect(self, s):
        self._check()
        return super(LLDPDU, self).post_dissect(s)

    def do_build(self):
        self._check()
        return super(LLDPDU, self).do_build()


def _ldp_id_adjustlen(pkt, x):
    """Return the length of the `id` field,
    according to its real encoded type"""
    f, v = pkt.getfield_and_val('id')
    length = f.i2len(pkt, v) + 1
    if (isinstance(pkt, LLDPDUPortID) and pkt.subtype == 0x4) or \
            (isinstance(pkt, LLDPDUChassisID) and pkt.subtype == 0x5):
        # Take the ConditionalField into account
        length += 1
    return length


def _ldp_id_lengthfrom(pkt):
    length = pkt._length
    if length is None:
        return 0
    # Subtract the subtype field
    length -= 1
    if (isinstance(pkt, LLDPDUPortID) and pkt.subtype == 0x4) or \
            (isinstance(pkt, LLDPDUChassisID) and pkt.subtype == 0x5):
        # Take the ConditionalField into account
        length -= 1
    return length


class LLDPDUChassisID(LLDPDU):
    """
        ieee 802.1ab-2016 - sec. 8.5.2 / p. 26
    """
    LLDP_CHASSIS_ID_TLV_SUBTYPES = {
        0x00: 'reserved',
        0x01: 'chassis component',
        0x02: 'interface alias',
        0x03: 'port component',
        0x04: 'MAC address',
        0x05: 'network address',
        0x06: 'interface name',
        0x07: 'locally assigned',
    }

    SUBTYPE_RESERVED = 0x00
    SUBTYPE_CHASSIS_COMPONENT = 0x01
    SUBTYPE_INTERFACE_ALIAS = 0x02
    SUBTYPE_PORT_COMPONENT = 0x03
    SUBTYPE_MAC_ADDRESS = 0x04
    SUBTYPE_NETWORK_ADDRESS = 0x05
    SUBTYPE_INTERFACE_NAME = 0x06
    SUBTYPE_LOCALLY_ASSIGNED = 0x07

    fields_desc = [
        BitEnumField('_type', 0x01, 7, LLDPDU.TYPES),
        BitFieldLenField('_length', None, 9, length_of='id',
                         adjust=lambda pkt, x: _ldp_id_adjustlen(pkt, x)),
        ByteEnumField('subtype', 0x00, LLDP_CHASSIS_ID_TLV_SUBTYPES),
        ConditionalField(
            ByteEnumField('family', 0, LLDPDU.IANA_ADDRESS_FAMILY_NUMBERS),
            lambda pkt: pkt.subtype == 0x05
        ),
        MultipleTypeField([
            (
                MACField('id', None),
                lambda pkt: pkt.subtype == 0x04
            ),
            (
                IPField('id', None),
                lambda pkt: pkt.subtype == 0x05 and pkt.family == 0x01
            ),
            (
                IP6Field('id', None),
                lambda pkt: pkt.subtype == 0x05 and pkt.family == 0x02
            ),
        ], StrLenField('id', '', length_from=_ldp_id_lengthfrom)
        )
    ]

    def _check(self):
        """
        run layer specific checks
        """
        if conf.contribs['LLDP'].strict_mode() and not self.id:
            raise LLDPInvalidLengthField('id must be >= 1 characters long')


class LLDPDUPortID(LLDPDU):
    """
        ieee 802.1ab-2016 - sec. 8.5.3 / p. 26
    """
    LLDP_PORT_ID_TLV_SUBTYPES = {
        0x00: 'reserved',
        0x01: 'interface alias',
        0x02: 'port component',
        0x03: 'MAC address',
        0x04: 'network address',
        0x05: 'interface name',
        0x06: 'agent circuit ID',
        0x07: 'locally assigned',
    }

    SUBTYPE_RESERVED = 0x00
    SUBTYPE_INTERFACE_ALIAS = 0x01
    SUBTYPE_PORT_COMPONENT = 0x02
    SUBTYPE_MAC_ADDRESS = 0x03
    SUBTYPE_NETWORK_ADDRESS = 0x04
    SUBTYPE_INTERFACE_NAME = 0x05
    SUBTYPE_AGENT_CIRCUIT_ID = 0x06
    SUBTYPE_LOCALLY_ASSIGNED = 0x07

    fields_desc = [
        BitEnumField('_type', 0x02, 7, LLDPDU.TYPES),
        BitFieldLenField('_length', None, 9, length_of='id',
                         adjust=lambda pkt, x: _ldp_id_adjustlen(pkt, x)),
        ByteEnumField('subtype', 0x00, LLDP_PORT_ID_TLV_SUBTYPES),
        ConditionalField(
            ByteEnumField('family', 0, LLDPDU.IANA_ADDRESS_FAMILY_NUMBERS),
            lambda pkt: pkt.subtype == 0x04
        ),
        MultipleTypeField([
            (
                MACField('id', None),
                lambda pkt: pkt.subtype == 0x03
            ),
            (
                IPField('id', None),
                lambda pkt: pkt.subtype == 0x04 and pkt.family == 0x01
            ),
            (
                IP6Field('id', None),
                lambda pkt: pkt.subtype == 0x04 and pkt.family == 0x02
            ),
        ], StrLenField('id', '', length_from=_ldp_id_lengthfrom)
        )
    ]

    def _check(self):
        """
        run layer specific checks
        """
        if conf.contribs['LLDP'].strict_mode() and not self.id:
            raise LLDPInvalidLengthField('id must be >= 1 characters long')


class LLDPDUTimeToLive(LLDPDU):
    """
        ieee 802.1ab-2016 - sec. 8.5.4 / p. 29
    """
    fields_desc = [
        BitEnumField('_type', 0x03, 7, LLDPDU.TYPES),
        BitField('_length', 0x02, 9),
        ShortField('ttl', 20)
    ]

    def _check(self):
        """
        run layer specific checks
        """
        if conf.contribs['LLDP'].strict_mode() and self._length != 2:
            raise LLDPInvalidLengthField('length must be 2 - got '
                                         '{}'.format(self._length))


class LLDPDUEndOfLLDPDU(LLDPDU):
    """
        ieee 802.1ab-2016 - sec. 8.5.1 / p. 26
    """
    fields_desc = [
        BitEnumField('_type', 0x00, 7, LLDPDU.TYPES),
        BitField('_length', 0x00, 9),
    ]

    def extract_padding(self, s):
        return '', s

    def _check(self):
        """
        run layer specific checks
        """
        if conf.contribs['LLDP'].strict_mode() and self._length != 0:
            raise LLDPInvalidLengthField('length must be 0 - got '
                                         '{}'.format(self._length))


class LLDPDUPortDescription(LLDPDU):
    """
        ieee 802.1ab-2016 - sec. 8.5.5 / p. 29
    """
    fields_desc = [
        BitEnumField('_type', 0x04, 7, LLDPDU.TYPES),
        BitFieldLenField('_length', None, 9, length_of='description'),
        StrLenField('description', '', length_from=lambda pkt: pkt._length)
    ]


class LLDPDUSystemName(LLDPDU):
    """
        ieee 802.1ab-2016 - sec. 8.5.6 / p. 30
    """
    fields_desc = [
        BitEnumField('_type', 0x05, 7, LLDPDU.TYPES),
        BitFieldLenField('_length', None, 9, length_of='system_name'),
        StrLenField('system_name', '', length_from=lambda pkt: pkt._length)
    ]


class LLDPDUSystemDescription(LLDPDU):
    """
        ieee 802.1ab-2016 - sec. 8.5.7 / p. 31
    """
    fields_desc = [
        BitEnumField('_type', 0x06, 7, LLDPDU.TYPES),
        BitFieldLenField('_length', None, 9, length_of='description'),
        StrLenField('description', '', length_from=lambda pkt: pkt._length)
    ]


class LLDPDUSystemCapabilities(LLDPDU):
    """
        ieee 802.1ab-2016 - sec. 8.5.8 / p. 31
    """
    fields_desc = [
        BitEnumField('_type', 0x07, 7, LLDPDU.TYPES),
        BitFieldLenField('_length', 4, 9),
        BitField('reserved_5_available', 0, 1),
        BitField('reserved_4_available', 0, 1),
        BitField('reserved_3_available', 0, 1),
        BitField('reserved_2_available', 0, 1),
        BitField('reserved_1_available', 0, 1),
        BitField('two_port_mac_relay_available', 0, 1),
        BitField('s_vlan_component_available', 0, 1),
        BitField('c_vlan_component_available', 0, 1),
        BitField('station_only_available', 0, 1),
        BitField('docsis_cable_device_available', 0, 1),
        BitField('telephone_available', 0, 1),
        BitField('router_available', 0, 1),
        BitField('wlan_access_point_available', 0, 1),
        BitField('mac_bridge_available', 0, 1),
        BitField('repeater_available', 0, 1),
        BitField('other_available', 0, 1),
        BitField('reserved_5_enabled', 0, 1),
        BitField('reserved_4_enabled', 0, 1),
        BitField('reserved_3_enabled', 0, 1),
        BitField('reserved_2_enabled', 0, 1),
        BitField('reserved_1_enabled', 0, 1),
        BitField('two_port_mac_relay_enabled', 0, 1),
        BitField('s_vlan_component_enabled', 0, 1),
        BitField('c_vlan_component_enabled', 0, 1),
        BitField('station_only_enabled', 0, 1),
        BitField('docsis_cable_device_enabled', 0, 1),
        BitField('telephone_enabled', 0, 1),
        BitField('router_enabled', 0, 1),
        BitField('wlan_access_point_enabled', 0, 1),
        BitField('mac_bridge_enabled', 0, 1),
        BitField('repeater_enabled', 0, 1),
        BitField('other_enabled', 0, 1),
    ]

    def _check(self):
        """
        run layer specific checks
        """
        if conf.contribs['LLDP'].strict_mode() and self._length != 4:
            raise LLDPInvalidLengthField('length must be 4 - got '
                                         '{}'.format(self._length))


class LLDPDUManagementAddress(LLDPDU):
    """
    ieee 802.1ab-2016 - sec. 8.5.9 / p. 32

    currently only 0x00..0x1e are used by standards, no way to
    use anything > 0xff as management address subtype is only
    one octet wide

    see https://www.iana.org/assignments/address-family-numbers/address-family-numbers.xhtml  # noqa: E501
    """

    SUBTYPE_MANAGEMENT_ADDRESS_OTHER = 0x00
    SUBTYPE_MANAGEMENT_ADDRESS_IPV4 = 0x01
    SUBTYPE_MANAGEMENT_ADDRESS_IPV6 = 0x02
    SUBTYPE_MANAGEMENT_ADDRESS_NSAP = 0x03
    SUBTYPE_MANAGEMENT_ADDRESS_HDLC = 0x04
    SUBTYPE_MANAGEMENT_ADDRESS_BBN = 0x05
    SUBTYPE_MANAGEMENT_ADDRESS_802 = 0x06
    SUBTYPE_MANAGEMENT_ADDRESS_E_163 = 0x07
    SUBTYPE_MANAGEMENT_ADDRESS_E_164 = 0x08
    SUBTYPE_MANAGEMENT_ADDRESS_F_69 = 0x09
    SUBTYPE_MANAGEMENT_ADDRESS_X_121 = 0x0A
    SUBTYPE_MANAGEMENT_ADDRESS_IPX = 0x0B
    SUBTYPE_MANAGEMENT_ADDRESS_APPLETALK = 0x0C
    SUBTYPE_MANAGEMENT_ADDRESS_DECNET_IV = 0x0D
    SUBTYPE_MANAGEMENT_ADDRESS_BANYAN_VINES = 0x0E
    SUBTYPE_MANAGEMENT_ADDRESS_E_164_WITH_NSAP = 0x0F
    SUBTYPE_MANAGEMENT_ADDRESS_DNS = 0x10
    SUBTYPE_MANAGEMENT_ADDRESS_DISTINGUISHED_NAME = 0x11
    SUBTYPE_MANAGEMENT_ADDRESS_AS_NUMBER = 0x12
    SUBTYPE_MANAGEMENT_ADDRESS_XTP_OVER_IPV4 = 0x13
    SUBTYPE_MANAGEMENT_ADDRESS_XTP_OVER_IPV6 = 0x14
    SUBTYPE_MANAGEMENT_ADDRESS_XTP_NATIVE_MODE_XTP = 0x15
    SUBTYPE_MANAGEMENT_ADDRESS_FIBER_CHANNEL_WORLD_WIDE_PORT_NAME = 0x16
    SUBTYPE_MANAGEMENT_ADDRESS_FIBER_CHANNEL_WORLD_WIDE_NODE_NAME = 0x17
    SUBTYPE_MANAGEMENT_ADDRESS_GWID = 0x18
    SUBTYPE_MANAGEMENT_ADDRESS_AFI_FOR_L2VPN = 0x19
    SUBTYPE_MANAGEMENT_ADDRESS_MPLS_TP_SECTION_ENDPOINT_ID = 0x1A
    SUBTYPE_MANAGEMENT_ADDRESS_MPLS_TP_LSP_ENDPOINT_ID = 0x1B
    SUBTYPE_MANAGEMENT_ADDRESS_MPLS_TP_PSEUDOWIRE_ENDPOINT_ID = 0x1C
    SUBTYPE_MANAGEMENT_ADDRESS_MT_IP_MULTI_TOPOLOGY_IPV4 = 0x1D
    SUBTYPE_MANAGEMENT_ADDRESS_MT_IP_MULTI_TOPOLOGY_IPV6 = 0x1E

    INTERFACE_NUMBERING_SUBTYPES = {
        0x01: 'unknown',
        0x02: 'ifIndex',
        0x03: 'system port number'
    }

    SUBTYPE_INTERFACE_NUMBER_UNKNOWN = 0x01
    SUBTYPE_INTERFACE_NUMBER_IF_INDEX = 0x02
    SUBTYPE_INTERFACE_NUMBER_SYSTEM_PORT_NUMBER = 0x03

    '''
    Note - calculation of _length field::

        _length = 1@_management_address_string_length +
                  1@management_address_subtype +
                  management_address.len +
                  1@interface_numbering_subtype +
                  4@interface_number +
                  1@_oid_string_length +
                  object_id.len
    '''

    fields_desc = [
        BitEnumField('_type', 0x08, 7, LLDPDU.TYPES),
        BitFieldLenField('_length', None, 9, length_of='management_address',
                         adjust=lambda pkt, x:
                         8 + len(pkt.management_address) + len(pkt.object_id)),
        BitFieldLenField('_management_address_string_length', None, 8,
                         length_of='management_address',
                         adjust=lambda pkt, x: len(pkt.management_address) + 1),  # noqa: E501
        ByteEnumField('management_address_subtype', 0x00,
                      LLDPDU.IANA_ADDRESS_FAMILY_NUMBERS),
        XStrLenField('management_address', '',
                     length_from=lambda pkt: 0
                     if pkt._management_address_string_length is None else
                     pkt._management_address_string_length - 1),
        ByteEnumField('interface_numbering_subtype',
                      SUBTYPE_INTERFACE_NUMBER_UNKNOWN,
                      INTERFACE_NUMBERING_SUBTYPES),
        BitField('interface_number', 0, 32),
        BitFieldLenField('_oid_string_length', None, 8, length_of='object_id'),
        XStrLenField('object_id', '',
                     length_from=lambda pkt: pkt._oid_string_length),
    ]

    def _check(self):
        """
        run layer specific checks
        """
        if conf.contribs['LLDP'].strict_mode():
            management_address_len = len(self.management_address)
            if management_address_len == 0 or management_address_len > 31:
                raise LLDPInvalidLengthField(
                    'management address must be  1..31 characters long - '
                    'got string of size {}'.format(management_address_len))


class ThreeBytesEnumField(EnumField, ThreeBytesField):

    def __init__(self, name, default, enum):
        EnumField.__init__(self, name, default, enum, "!I")


class LLDPDUGenericOrganisationSpecific(LLDPDU):

    ORG_UNIQUE_CODE_PNO = 0x000ecf
    ORG_UNIQUE_CODE_IEEE_802_1 = 0x0080c2
    ORG_UNIQUE_CODE_IEEE_802_3 = 0x00120f
    ORG_UNIQUE_CODE_TIA_TR_41_MED = 0x0012bb
    ORG_UNIQUE_CODE_HYTEC = 0x30b216

    ORG_UNIQUE_CODES = {
        ORG_UNIQUE_CODE_PNO: "PROFIBUS International (PNO)",
        ORG_UNIQUE_CODE_IEEE_802_1: "IEEE 802.1",
        ORG_UNIQUE_CODE_IEEE_802_3: "IEEE 802.3",
        ORG_UNIQUE_CODE_TIA_TR_41_MED: "TIA TR-41 Committee . Media Endpoint Discovery",  # noqa: E501
        ORG_UNIQUE_CODE_HYTEC: "Hytec Geraetebau GmbH"
    }

    fields_desc = [
        BitEnumField('_type', 127, 7, LLDPDU.TYPES),
        BitFieldLenField('_length', None, 9, length_of='data', adjust=lambda pkt, x: len(pkt.data) + 4),  # noqa: E501
        ThreeBytesEnumField('org_code', 0, ORG_UNIQUE_CODES),
        ByteField('subtype', 0x00),
        XStrLenField('data', '',
                     length_from=lambda pkt: 0 if pkt._length is None else
                     pkt._length - 4)
    ]

    @staticmethod
    def _match_organization_specific(payload):
        return True


class LLDPDUPowerViaMDI(LLDPDUGenericOrganisationSpecific):
    """
    Legacy PoE TLV originally defined in IEEE Std 802.1AB-2005 Annex G.3.

    IEEE802.3bt-2018 - sec. 79.3.2.
    """

    # IEEE802.3bt-2018 - sec. 79.3.2.1
    MDI_POWER_SUPPORT = {
        (1 << 3): 'PSE pairs controlled',
        (1 << 2): 'PSE MDI power enabled',
        (1 << 1): 'PSE MDI power supported',
        (1 << 0): 'port class PSE',
    }

    # IEEE802.3bt-2018 - sec. 79.3.2.2
    PSE_POWER_PAIR = {
        1: 'alt A',
        2: 'alt B',
    }

    # IEEE802.3bt-2018 - sec. 79.3.2.3
    POWER_CLASS = {
        1: 'class 0',
        2: 'class 1',
        3: 'class 2',
        4: 'class 3',
        5: 'class 4 and above',
    }

    fields_desc = [
        BitEnumField('_type', 127, 7, LLDPDU.TYPES),
        BitField('_length', 7, 9),
        ThreeBytesField('org_code', LLDPDUGenericOrganisationSpecific.ORG_UNIQUE_CODE_IEEE_802_3),  # noqa: E501
        ByteField('subtype', 2),
        FlagsField('MDI_power_support', 0, 8, MDI_POWER_SUPPORT),
        ByteEnumField('PSE_power_pair', 1, PSE_POWER_PAIR),
        ByteEnumField('power_class', 1, POWER_CLASS),
    ]

    @staticmethod
    def _match_organization_specific(payload):
        """
        match organization specific TLV
        """
        return (orb(payload[5]) == 2 and orb(payload[1]) == 7
                and bytes_int(payload[2:5]) ==
                LLDPDUGenericOrganisationSpecific.ORG_UNIQUE_CODE_IEEE_802_3)

    def _check(self):
        """
        run layer specific checks
        """
        if conf.contribs['LLDP'].strict_mode() and self._length != 7:
            raise LLDPInvalidLengthField('length must be 7 - got '
                                         '{}'.format(self._length))


class LLDPDUPowerViaMDIDDL(LLDPDUPowerViaMDI):
    """
    PoE TLV with DLL classification extension specified in IEEE802.3at-2009

    Note: power values are expressed in units of Watts,
    converted to tenth of Watts internally

    IEEE802.3bt-2018 - sec. 79.3.2
    """

    # IEEE802.3bt-2018 - sec. 79.3.2.4
    POWER_TYPE_NO = {
        1: 'type 1',
        0: 'type 2',
    }

    # IEEE802.3bt-2018 - sec. 79.3.2.4
    POWER_TYPE_DIR = {
        1: 'PD',
        0: 'PSE',
    }

    # IEEE802.3bt-2018 - sec. 79.3.2.4
    POWER_SOURCE_PD = {
        0b11: 'PSE and local',
        0b10: 'reserved',
        0b01: 'PSE',
        0b00: 'unknown',
    }

    # IEEE802.3bt-2018 - sec. 79.3.2.4
    POWER_SOURCE_PSE = {
        0b11: 'reserved',
        0b10: 'backup source',
        0b01: 'primary source',
        0b00: 'unknown',
    }

    # IEEE802.3bt-2018 - sec. 79.3.2.4
    PD_4PID_SUP = {
        0: 'not supported',
        1: 'supported',
    }

    # IEEE802.3bt-2018 - sec. 79.3.2.4
    POWER_PRIO = {
        0b11: 'low',
        0b10: 'high',
        0b01: 'critical',
        0b00: 'unknown',
    }

    fields_desc = [
        BitEnumField('_type', 127, 7, LLDPDU.TYPES),
        BitField('_length', 12, 9),
        ThreeBytesField('org_code', LLDPDUGenericOrganisationSpecific.ORG_UNIQUE_CODE_IEEE_802_3),  # noqa: E501
        ByteField('subtype', 2),
        FlagsField('MDI_power_support', 0, 8, LLDPDUPowerViaMDI.MDI_POWER_SUPPORT),
        ByteEnumField('PSE_power_pair', 1, LLDPDUPowerViaMDI.PSE_POWER_PAIR),
        ByteEnumField('power_class', 1, LLDPDUPowerViaMDI.POWER_CLASS),
        BitEnumField('power_type_no', 1, 1, POWER_TYPE_NO),
        BitEnumField('power_type_dir', 1, 1, POWER_TYPE_DIR),
        MultipleTypeField([
            (
                BitEnumField('power_source', 0b01, 2, POWER_SOURCE_PD),
                lambda pkt: pkt.power_type_dir == 1
            ),
        ], BitEnumField('power_source', 0b01, 2, POWER_SOURCE_PSE)),
        MultipleTypeField([
            (
                BitEnumField('PD_4PID', 0, 2, PD_4PID_SUP),
                lambda pkt: pkt.power_type_dir == 1
            ),
        ], BitField('PD_4PID', 0, 2)),
        BitEnumField('power_prio', 0, 2, POWER_PRIO),
        ScalingField('PD_requested_power', 0, scaling=0.1,
                     unit='W', ndigits=1, fmt='H'),
        ScalingField('PSE_allocated_power', 0, scaling=0.1,
                     unit='W', ndigits=1, fmt='H'),
    ]

    @staticmethod
    def _match_organization_specific(payload):
        """
        match organization specific TLV
        """
        return (orb(payload[5]) == 2 and orb(payload[1]) == 12
                and bytes_int(payload[2:5]) ==
                LLDPDUGenericOrganisationSpecific.ORG_UNIQUE_CODE_IEEE_802_3)

    def _check(self):
        """
        run layer specific checks
        """
        if conf.contribs['LLDP'].strict_mode() and self._length != 12:
            raise LLDPInvalidLengthField('length must be 12 - got '
                                         '{}'.format(self._length))
        # IEEE802.3bt-2018 - sec. 79.3.2.{5,6}
        for field, description, max_value in [('PD_requested_power',
                                               'PSE requested power',
                                               99.9),
                                              ('PSE_allocated_power',
                                               'PSE allocated power',
                                               99.9)]:
            val = getattr(self, field)
            if (conf.contribs['LLDP'].strict_mode() and val > max_value):
                raise LLDPInvalidFieldValue(
                    'exceeded maximum {} of {} - got '
                    '{}'.format(description, max_value, val))


class LLDPDUPowerViaMDIType34(LLDPDUPowerViaMDIDDL):
    """
    PoE TLV with DLL classification and type 3 and 4 extensions
    specified in IEEE802.3bt-2018

    Note: power values are expressed in units of Watts,
    converted to tenth of Watts internally

    IEEE802.3bt-2018 - sec. 79.3.2
    """

    # IEEE802.3bt-2018 - sec. 79.3.2.6e
    PSE_POWERING_STATUS = {
        0b11: '4-pair powering dual-signature PD',
        0b10: '4-pair powering single-signature PD',
        0b01: '2-pair powering',
        0b00: 'ignore',
    }

    # IEEE802.3bt-2018 - sec. 79.3.2.6e
    PD_POWERED_STATUS = {
        0b11: '4-pair powered dual-signature PD',
        0b10: '2-pair powered dual-signature PD',
        0b01: 'powered single-signature PD',
        0b00: 'ignore',
    }

    # IEEE802.3bt-2018 - sec. 79.3.2.6e
    PSE_POWER_PAIRS_EXT = {
        0b11: 'both alts',
        0b10: 'alt A',
        0b01: 'alt B',
        0b00: 'ignore',
    }

    # IEEE802.3bt-2018 - sec. 79.3.2.6e
    DUAL_SIGNATURE_POWER_CLASS = {
        0b111: 'single-signature PD or 2-pair only PSE',
        0b110: 'ignore',
        0b101: 'class 5',
        0b100: 'class 4',
        0b011: 'class 3',
        0b010: 'class 2',
        0b001: 'class 1',
        0b000: 'ignore',
    }

    # IEEE802.3bt-2018 - sec. 79.3.2.6e
    POWER_CLASS_EXT = {
        0b1111: 'dual-signature pd',
        0b1110: 'ignore',
        0b1101: 'ignore',
        0b1100: 'ignore',
        0b1011: 'ignore',
        0b1010: 'ignore',
        0b1001: 'ignore',
        0b1000: 'class 8',
        0b0111: 'class 7',
        0b0110: 'class 6',
        0b0101: 'class 5',
        0b0100: 'class 4',
        0b0011: 'class 3',
        0b0010: 'class 2',
        0b0001: 'class 1',
        0b0000: 'ignore',
    }

    # IEEE802.3bt-2018 - sec. 79.3.2.6d
    POWER_TYPE_EXT = {
        0b111: 'ignore',
        0b110: 'ignore',
        0b101: 'type 4 dual-signature PD',
        0b100: 'type 4 single-signature PD',
        0b011: 'type 3 dual-signature PD',
        0b010: 'type 3 single-signature PD',
        0b001: 'type 4 PSE',
        0b000: 'type 3 PSE',
    }

    # IEEE802.3bt-2018 - sec. 79.3.2.6d
    PD_LOAD = {
        1: 'dual-signature and electrically isolated',
        0: 'single-signature or not electrically isolated',
    }

    # IEEE802.3bt-2018 - sec. 79.3.2.6h
    AUTOCLASS = {
        (1 << 2): 'PSE autoclass support',
        (1 << 1): 'autoclass completed',
        (1 << 0): 'autoclass request',
    }

    # IEEE802.3bt-2018 - sec. 79.3.2.6i
    POWER_DOWN_REQ = {
        0x1d: 'power down',
        0: 'ignore',
    }

    fields_desc = [
        BitEnumField('_type', 127, 7, LLDPDU.TYPES),
        BitField('_length', 29, 9),
        ThreeBytesField('org_code', LLDPDUGenericOrganisationSpecific.ORG_UNIQUE_CODE_IEEE_802_3),  # noqa: E501
        ByteField('subtype', 2),
        FlagsField('MDI_power_support', 0, 8, LLDPDUPowerViaMDI.MDI_POWER_SUPPORT),
        ByteEnumField('PSE_power_pair', 1, LLDPDUPowerViaMDI.PSE_POWER_PAIR),
        ByteEnumField('power_class', 1, LLDPDUPowerViaMDI.POWER_CLASS),
        BitEnumField('power_type_no', 1, 1, LLDPDUPowerViaMDIDDL.POWER_TYPE_NO),
        BitEnumField('power_type_dir', 1, 1, LLDPDUPowerViaMDIDDL.POWER_TYPE_DIR),
        MultipleTypeField([
            (
                BitEnumField('power_source', 0b01, 2, LLDPDUPowerViaMDIDDL.POWER_SOURCE_PD),  # noqa: E501
                lambda pkt: pkt.power_type_dir == 1
            ),
        ], BitEnumField('power_source', 0b01, 2, LLDPDUPowerViaMDIDDL.POWER_SOURCE_PSE)),  # noqa: E501
        MultipleTypeField([
            (
                BitEnumField('PD_4PID', 0, 2, LLDPDUPowerViaMDIDDL.PD_4PID_SUP),
                lambda pkt: pkt.power_type_dir == 1
            ),
        ], BitField('PD_4PID', 0, 2)),
        BitEnumField('power_prio', 0, 2, LLDPDUPowerViaMDIDDL.POWER_PRIO),
        ScalingField('PD_requested_power', 0, scaling=0.1,
                     unit='W', ndigits=1, fmt='H'),
        ScalingField('PSE_allocated_power', 0, scaling=0.1,
                     unit='W', ndigits=1, fmt='H'),
        ScalingField('PD_requested_power_mode_A', 0, scaling=0.1,
                     unit='W', ndigits=1, fmt='H'),
        ScalingField('PD_requested_power_mode_B', 0, scaling=0.1,
                     unit='W', ndigits=1, fmt='H'),
        ScalingField('PD_allocated_power_alt_A', 0, scaling=0.1,
                     unit='W', ndigits=1, fmt='H'),
        ScalingField('PD_allocated_power_alt_B', 0, scaling=0.1,
                     unit='W', ndigits=1, fmt='H'),
        BitEnumField('PSE_powering_status', 0, 2, PSE_POWERING_STATUS),
        BitEnumField('PD_powered_status', 0, 2, PD_POWERED_STATUS),
        BitEnumField('PD_power_pair_ext', 0, 2, PSE_POWER_PAIRS_EXT),
        BitEnumField('dual_signature_class_mode_A',
                     0b111, 3, DUAL_SIGNATURE_POWER_CLASS),
        BitEnumField('dual_signature_class_mode_B',
                     0b111, 3, DUAL_SIGNATURE_POWER_CLASS),
        BitEnumField('power_class_ext', 0, 4, POWER_CLASS_EXT),
        BitEnumField('power_type_ext', 0, 7, POWER_TYPE_EXT),
        BitEnumField('PD_load', 0, 1, PD_LOAD),
        ScalingField('PSE_max_available_power', 0, scaling=0.1,
                     unit='W', ndigits=1, fmt='H'),
        FlagsField('autoclass', 0, 8, AUTOCLASS),
        BitEnumField('power_down_req', 0, 6, POWER_DOWN_REQ),
        BitScalingField('power_down_time', 0, 18, unit='s'),
    ]

    @staticmethod
    def _match_organization_specific(payload):
        '''
        match organization specific TLV
        '''
        return (orb(payload[5]) == 2 and orb(payload[1]) == 29
                and bytes_int(payload[2:5]) ==
                LLDPDUGenericOrganisationSpecific.ORG_UNIQUE_CODE_IEEE_802_3)

    def _check(self):
        """
        run layer specific checks
        """
        if conf.contribs['LLDP'].strict_mode() and self._length != 29:
            raise LLDPInvalidLengthField('length must be 29 - got '
                                         '{}'.format(self._length))
        # IEEE802.3bt-2018 - sec. 79.3.2.6{a..b,e,g}
        for field, description, max_value in [('PD_requested_power',
                                               'PSE requested power',
                                               99.9),
                                              ('PSE_allocated_power',
                                               'PSE allocated power',
                                               99.9),
                                              ('PD_requested_power_mode_A',
                                               'PD requested power mode A',
                                               49.9),
                                              ('PD_requested_power_mode_B',
                                               'PD requested power mode B',
                                               49.9),
                                              ('PD_allocated_power_alt_A',
                                               'PD allocated power alt A',
                                               49.9),
                                              ('PD_allocated_power_alt_B',
                                               'PD allocated power alt B',
                                               49.9),
                                              ('PSE_max_available_power',
                                               'PSE maximum available power',
                                               99.9),
                                              ('power_down_time',
                                               'power down time',
                                               262143)]:
            val = getattr(self, field) or 0
            if (conf.contribs['LLDP'].strict_mode() and val > max_value):
                raise LLDPInvalidFieldValue(
                    'exceeded maximum {} of {} - got '
                    '{}'.format(description, max_value, val))


class LLDPDUPowerViaMDIMeasure(LLDPDUGenericOrganisationSpecific):
    """
    PoE TLV measurements in IEEE802.3bt-2018

    Note: power values are expressed in units of Watts,
    converted to hundredths of Watts internally;
    energy values are expressed in units of Joules,
    converted to tenths of kilo-Joules internally;
    voltage values are expressed in units of Volts,
    converted to milli-Volts internally;
    current values are expressed in units of Amperes,
    converted to tenths of milli-Amperes internally.
    PSE price index is converted internally.

    IEEE802.3bt-2018 - sec. 79.3.8
    """

    MEASURE_TYPE = {
        (1 << 3): 'voltage',
        (1 << 2): 'current',
        (1 << 1): 'power',
        (1 << 0): 'energy',
    }

    MEASURE_SOURCE = {
        0b00: 'no request',
        0b01: 'mode A',
        0b10: 'mode B',
        0b11: 'port total',
    }

    POWER_PRICE_INDEX = {
        0xffff: 'not available',
    }

    @staticmethod
    def _encode_ppi(val):
        # IEEE802.3bt-2018 - sec. 79.3.8
        return int(75046 / 2.512 * (val ** (1 / 5)) - 10046)

    @staticmethod
    def _decode_ppi(val):
        # IEEE802.3bt-2018 - sec. 79.3.8
        return ((val + 10046) * 2.512 / 75046) ** 5

    fields_desc = [
        BitEnumField('_type', 127, 7, LLDPDU.TYPES),
        BitField('_length', 26, 9),
        ThreeBytesField('org_code', LLDPDUGenericOrganisationSpecific.ORG_UNIQUE_CODE_IEEE_802_3),  # noqa: E501
        ByteField('subtype', 8),
        FlagsField('support', 0, 4, MEASURE_TYPE),
        BitEnumField('source', 0, 4, MEASURE_SOURCE),
        FlagsField('request', 0, 4, MEASURE_TYPE),
        FlagsField('valid', 0, 4, MEASURE_TYPE),
        ScalingField('voltage_uncertainty', 0, scaling=0.001,
                     unit='V', ndigits=3, fmt='H'),
        ScalingField('current_uncertainty', 0, scaling=0.0001,
                     unit='A', ndigits=4, fmt='H'),
        ScalingField('power_uncertainty', 0, scaling=0.01,
                     unit='W', ndigits=2, fmt='H'),
        ScalingField('energy_uncertainty', 0, scaling=100,
                     unit='J', ndigits=0, fmt='H'),
        ScalingField('voltage_measurement', 0, scaling=0.001,
                     unit='V', ndigits=3, fmt='H'),
        ScalingField('current_measurement', 0, scaling=0.0001,
                     unit='A', ndigits=4, fmt='H'),
        ScalingField('power_measurement', 0, scaling=0.01,
                     unit='W', ndigits=2, fmt='H'),
        ScalingField('energy_measurement', 0, scaling=100,
                     unit='J', ndigits=0, fmt='I'),
        ShortEnumField('power_price_index', 0xffff, POWER_PRICE_INDEX),
    ]

    def do_build(self):
        backup_ppi = self.power_price_index
        self.power_price_index = 0xffff if self.power_price_index == 0xffff \
            else LLDPDUPowerViaMDIMeasure._encode_ppi(self.power_price_index)
        s = super(LLDPDUPowerViaMDIMeasure, self).do_build()
        self.power_price_index = backup_ppi
        return s

    def post_dissect(self, s):
        s = super(LLDPDUPowerViaMDIMeasure, self).post_dissect(s)
        self.power_price_index = 0xffff if self.power_price_index == 0xffff \
            else LLDPDUPowerViaMDIMeasure._decode_ppi(self.power_price_index)
        return s

    @staticmethod
    def _match_organization_specific(payload):
        '''
        match organization specific TLV
        '''
        return (orb(payload[5]) == 8 and orb(payload[1]) == 26
                and bytes_int(payload[2:5]) ==
                LLDPDUGenericOrganisationSpecific.ORG_UNIQUE_CODE_IEEE_802_3)

    def _check(self):
        """
        run layer specific checks
        """
        if conf.contribs['LLDP'].strict_mode() and self._length != 26:
            raise LLDPInvalidLengthField('length must be 26 - got '
                                         '{}'.format(self._length))
        # IEEE802.3bt-2018 - sec. 79.3.8
        for field, description, max_value in [('voltage_uncertainty',
                                               'voltage uncertainty',
                                               65),
                                              ('voltage_measurement',
                                               'voltage measurement',
                                               65),
                                              ('current_uncertainty',
                                               'current uncertainty',
                                               6.5),
                                              ('current_measurement',
                                               'current measurement',
                                               6.5),
                                              ('energy_uncertainty',
                                               'energy uncertainty',
                                               6500000),
                                              ('power_uncertainty',
                                               'power uncertainty',
                                               650),
                                              ('power_measurement',
                                               'power measurement',
                                               650)]:
            val = getattr(self, field) or 0
            if (conf.contribs['LLDP'].strict_mode() and val > max_value):
                raise LLDPInvalidFieldValue(
                    'exceeded maximum {} of {} - got '
                    '{}'.format(description, max_value, val))
            val = self.power_price_index or 0xffff
            if val > 65000 and val != 0xffff:
                raise LLDPInvalidFieldValue(
                    'exceeded maximum power price index of {} - got '
                    '{}'.format(LLDPDUPowerViaMDIMeasure._decode_ppi(65000),
                                LLDPDUPowerViaMDIMeasure._decode_ppi(val)))


# 0x09 .. 0x7e is reserved for future standardization and for now treated as Raw() data  # noqa: E501
LLDPDU_CLASS_TYPES = {
    0x00: LLDPDUEndOfLLDPDU,
    0x01: LLDPDUChassisID,
    0x02: LLDPDUPortID,
    0x03: LLDPDUTimeToLive,
    0x04: LLDPDUPortDescription,
    0x05: LLDPDUSystemName,
    0x06: LLDPDUSystemDescription,
    0x07: LLDPDUSystemCapabilities,
    0x08: LLDPDUManagementAddress,
    127: [
        LLDPDUPowerViaMDI,
        LLDPDUPowerViaMDIDDL,
        LLDPDUPowerViaMDIType34,
        LLDPDUPowerViaMDIMeasure,
        LLDPDUGenericOrganisationSpecific,
    ]
}


class LLDPConfiguration(object):
    """
    basic configuration for LLDP layer
    """

    def __init__(self):
        self._strict_mode = True
        self.strict_mode_enable()

    def strict_mode_enable(self):
        """
        enable strict mode and dissector debugging
        """
        self._strict_mode = True

    def strict_mode_disable(self):
        """
        disable strict mode and dissector debugging
        """
        self._strict_mode = False

    def strict_mode(self):
        """
        get current strict mode state
        """
        return self._strict_mode


conf.contribs['LLDP'] = LLDPConfiguration()

bind_layers(Ether, LLDPDU, type=LLDP_ETHER_TYPE)
bind_layers(Dot1Q, LLDPDU, type=LLDP_ETHER_TYPE)



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information

# scapy.contrib.description = TaZmen Sniffer Protocol (TZSP)
# scapy.contrib.status = loads

"""
    TZSP - TaZmen Sniffer Protocol
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :author:    Thomas Tannhaeuser, hecke@naberius.de

    :description:

        This module provides Scapy layers for the TZSP protocol.

        references:
            - https://en.wikipedia.org/wiki/TZSP
            - https://web.archive.org/web/20050404125022/http://www.networkchemistry.com/support/appnotes/an001_tzsp.html  # noqa: E501

    :NOTES:
        - to allow Scapy to dissect this layer automatically, you need to bind the TZSP layer to UDP using  # noqa: E501
          the default TZSP port (0x9090), e.g.

            bind_layers(UDP, TZSP, sport=TZSP_PORT_DEFAULT)
            bind_layers(UDP, TZSP, dport=TZSP_PORT_DEFAULT)

        - packet format definition from www.networkchemistry.com is different from the one given by wikipedia  # noqa: E501
        - seems Wireshark implements the wikipedia protocol version (didn't dive into their code)  # noqa: E501
        - observed (miss)behavior of Wireshark (2.2.6)
          - fails to decode RSSI & SNR using short values - only one byte taken
          - SNR is labeled as silence
          - WlanRadioHdrSerial is labeled as Sensor MAC
          - doesn't know the packet count tag (40 / 0x28)

"""
from scapy.compat import orb
from scapy.contrib.avs import AVSWLANHeader
from scapy.error import warning, Scapy_Exception
from scapy.fields import ByteField, ShortEnumField, IntField, FieldLenField, YesNoByteField  # noqa: E501
from scapy.layers.dot11 import Packet, Dot11, PrismHeader
from scapy.layers.l2 import Ether
from scapy.fields import StrLenField, ByteEnumField, ShortField, XStrLenField
from scapy.packet import Raw


TZSP_PORT_DEFAULT = 0x9090


class TZSP(Packet):
    TYPE_RX_PACKET = 0x00
    TYPE_TX_PACKET = 0x01
    TYPE_CONFIG = 0x03
    TYPE_KEEPALIVE = TYPE_NULL = 0x04
    TYPE_PORT = 0x05

    TYPES = {
        TYPE_RX_PACKET: 'RX_PACKET',
        TYPE_TX_PACKET: 'TX_PACKET',
        TYPE_CONFIG: 'CONFIG',
        TYPE_NULL: 'KEEPALIVE/NULL',
        TYPE_PORT: 'PORT',
    }

    ENCAPSULATED_ETHERNET = 0x01
    ENCAPSULATED_IEEE_802_11 = 0x12
    ENCAPSULATED_PRISM_HEADER = 0x77
    ENCAPSULATED_WLAN_AVS = 0x7f

    ENCAPSULATED_PROTOCOLS = {
        ENCAPSULATED_ETHERNET: 'ETHERNET',
        ENCAPSULATED_IEEE_802_11: 'IEEE 802.11',
        ENCAPSULATED_PRISM_HEADER: 'PRISM HEADER',
        ENCAPSULATED_WLAN_AVS: 'WLAN AVS'
    }

    ENCAPSULATED_PROTOCOL_CLASSES = {
        ENCAPSULATED_ETHERNET: Ether,
        ENCAPSULATED_IEEE_802_11: Dot11,
        ENCAPSULATED_PRISM_HEADER: PrismHeader,
        ENCAPSULATED_WLAN_AVS: AVSWLANHeader
    }

    fields_desc = [
        ByteField('version', 0x01),
        ByteEnumField('type', TYPE_RX_PACKET, TYPES),
        ShortEnumField('encapsulated_protocol', ENCAPSULATED_ETHERNET, ENCAPSULATED_PROTOCOLS)  # noqa: E501
    ]

    def get_encapsulated_payload_class(self):
        """
        get the class that holds the encapsulated payload of the TZSP packet
        :return: class representing the payload, Raw() on error
        """

        try:
            return TZSP.ENCAPSULATED_PROTOCOL_CLASSES[self.encapsulated_protocol]  # noqa: E501
        except KeyError:
            warning(
                'unknown or invalid encapsulation type (%i) - returning payload as raw()' % self.encapsulated_protocol)  # noqa: E501
            return Raw

    def guess_payload_class(self, payload):
        if self.type == TZSP.TYPE_KEEPALIVE:
            if len(payload):
                warning('payload (%i bytes) in KEEPALIVE/NULL packet',
                        len(payload))
            return Raw
        else:
            return _tzsp_guess_next_tag(payload)

    def get_encapsulated_payload(self):

        has_encapsulated_data = self.type == TZSP.TYPE_RX_PACKET or self.type == TZSP.TYPE_TX_PACKET  # noqa: E501

        if has_encapsulated_data:
            end_tag_lyr = self.payload.getlayer(TZSPTagEnd)
            if end_tag_lyr:
                return end_tag_lyr.payload
            else:
                return None


def _tzsp_handle_unknown_tag(payload, tag_type):

    payload_len = len(payload)

    if payload_len < 2:
        warning('invalid or unknown tag type (%i) and too short packet - '
                'treat remaining data as Raw', tag_type)
        return Raw

    tag_data_length = orb(payload[1])

    tag_data_fits_in_payload = (tag_data_length + 2) <= payload_len
    if not tag_data_fits_in_payload:
        warning('invalid or unknown tag type (%i) and too short packet - '
                'treat remaining data as Raw', tag_type)
        return Raw

    warning('invalid or unknown tag type (%i)', tag_type)

    return TZSPTagUnknown


def _tzsp_guess_next_tag(payload):
    """
    :return: class representing the next tag, Raw on error, None on missing payload  # noqa: E501
    """

    if not payload:
        warning('missing payload')
        return None

    tag_type = orb(payload[0])

    try:
        tag_class_definition = _TZSP_TAG_CLASSES[tag_type]

    except KeyError:

        return _tzsp_handle_unknown_tag(payload, tag_type)

    if type(tag_class_definition) is not dict:
        return tag_class_definition

    try:
        length = orb(payload[1])
    except IndexError:
        length = None

    if not length:
        warning('no tag length given - packet too short')
        return Raw

    try:
        return tag_class_definition[length]
    except KeyError:
        warning('invalid tag length %s for tag type %s', length, tag_type)
        return Raw


class _TZSPTag(Packet):
    TAG_TYPE_PADDING = 0x00
    TAG_TYPE_END = 0x01
    TAG_TYPE_RAW_RSSI = 0x0a
    TAG_TYPE_SNR = 0x0b
    TAG_TYPE_DATA_RATE = 0x0c
    TAG_TYPE_TIMESTAMP = 0x0d
    TAG_TYPE_CONTENTION_FREE = 0x0f
    TAG_TYPE_DECRYPTED = 0x10
    TAG_TYPE_FCS_ERROR = 0x11
    TAG_TYPE_RX_CHANNEL = 0x12
    TAG_TYPE_PACKET_COUNT = 0x28
    TAG_TYPE_RX_FRAME_LENGTH = 0x29
    TAG_TYPE_WLAN_RADIO_HDR_SERIAL = 0x3c

    TAG_TYPES = {
        TAG_TYPE_PADDING: 'PADDING',
        TAG_TYPE_END: 'END',
        TAG_TYPE_RAW_RSSI: 'RAW_RSSI',
        TAG_TYPE_SNR: 'SNR',
        TAG_TYPE_DATA_RATE: 'DATA_RATE',
        TAG_TYPE_TIMESTAMP: 'TIMESTAMP',
        TAG_TYPE_CONTENTION_FREE: 'CONTENTION_FREE',
        TAG_TYPE_DECRYPTED: 'DECRYPTED',
        TAG_TYPE_FCS_ERROR: 'FCS_ERROR',
        TAG_TYPE_RX_CHANNEL: 'RX_CHANNEL',
        TAG_TYPE_PACKET_COUNT: 'PACKET_COUNT',
        TAG_TYPE_RX_FRAME_LENGTH: 'RX_FRAME_LENGTH',
        TAG_TYPE_WLAN_RADIO_HDR_SERIAL: 'WLAN_RADIO_HDR_SERIAL'
    }

    def guess_payload_class(self, payload):
        return _tzsp_guess_next_tag(payload)


class TZSPStructureException(Scapy_Exception):
    pass


class TZSPTagPadding(_TZSPTag):
    """
    padding tag (should be ignored)
    """
    fields_desc = [
        ByteEnumField('type', _TZSPTag.TAG_TYPE_PADDING, _TZSPTag.TAG_TYPES),
    ]


class TZSPTagEnd(Packet):
    """
    last tag
    """
    fields_desc = [
        ByteEnumField('type', _TZSPTag.TAG_TYPE_END, _TZSPTag.TAG_TYPES),
    ]

    def guess_payload_class(self, payload):
        """
        the type of the payload encapsulation is given be the outer TZSP layers attribute encapsulation_protocol  # noqa: E501
        """

        under_layer = self.underlayer
        tzsp_header = None

        while under_layer:
            if isinstance(under_layer, TZSP):
                tzsp_header = under_layer
                break
            under_layer = under_layer.underlayer

        if tzsp_header:

            return tzsp_header.get_encapsulated_payload_class()
        else:
            raise TZSPStructureException('missing parent TZSP header')


class TZSPTagRawRSSIByte(_TZSPTag):
    """
    relative received signal strength - signed byte value
    """
    fields_desc = [
        ByteEnumField('type', _TZSPTag.TAG_TYPE_RAW_RSSI, _TZSPTag.TAG_TYPES),
        ByteField('len', 1),
        ByteField('raw_rssi', 0)
    ]


class TZSPTagRawRSSIShort(_TZSPTag):
    """
    relative received signal strength - signed short value
    """
    fields_desc = [
        ByteEnumField('type', _TZSPTag.TAG_TYPE_RAW_RSSI, _TZSPTag.TAG_TYPES),
        ByteField('len', 2),
        ShortField('raw_rssi', 0)
    ]


class TZSPTagSNRByte(_TZSPTag):
    """
    signal noise ratio - signed byte value
    """
    fields_desc = [
        ByteEnumField('type', _TZSPTag.TAG_TYPE_SNR, _TZSPTag.TAG_TYPES),
        ByteField('len', 1),
        ByteField('snr', 0)
    ]


class TZSPTagSNRShort(_TZSPTag):
    """
    signal noise ratio - signed short value
    """
    fields_desc = [
        ByteEnumField('type', _TZSPTag.TAG_TYPE_SNR, _TZSPTag.TAG_TYPES),
        ByteField('len', 2),
        ShortField('snr', 0)
    ]


class TZSPTagDataRate(_TZSPTag):
    """
    wireless link data rate
    """
    DATA_RATE_UNKNOWN = 0x00
    DATA_RATE_1 = 0x02
    DATA_RATE_2 = 0x04
    DATA_RATE_5_5 = 0x0B
    DATA_RATE_6 = 0x0C
    DATA_RATE_9 = 0x12
    DATA_RATE_11 = 0x16
    DATA_RATE_12 = 0x18
    DATA_RATE_18 = 0x24
    DATA_RATE_22 = 0x2C
    DATA_RATE_24 = 0x30
    DATA_RATE_33 = 0x42
    DATA_RATE_36 = 0x48
    DATA_RATE_48 = 0x60
    DATA_RATE_54 = 0x6C
    DATA_RATE_LEGACY_1 = 0x0A
    DATA_RATE_LEGACY_2 = 0x14
    DATA_RATE_LEGACY_5_5 = 0x37
    DATA_RATE_LEGACY_11 = 0x6E

    DATA_RATES = {
        DATA_RATE_UNKNOWN: 'unknown',
        DATA_RATE_1: '1 MB/s',
        DATA_RATE_2: '2 MB/s',
        DATA_RATE_5_5: '5.5 MB/s',
        DATA_RATE_6: '6 MB/s',
        DATA_RATE_9: '9 MB/s',
        DATA_RATE_11: '11 MB/s',
        DATA_RATE_12: '12 MB/s',
        DATA_RATE_18: '18 MB/s',
        DATA_RATE_22: '22 MB/s',
        DATA_RATE_24: '24 MB/s',
        DATA_RATE_33: '33 MB/s',
        DATA_RATE_36: '36 MB/s',
        DATA_RATE_48: '48 MB/s',
        DATA_RATE_54: '54 MB/s',
        DATA_RATE_LEGACY_1: '1 MB/s (legacy)',
        DATA_RATE_LEGACY_2: '2 MB/s (legacy)',
        DATA_RATE_LEGACY_5_5: '5.5 MB/s (legacy)',
        DATA_RATE_LEGACY_11: '11 MB/s (legacy)',
    }

    fields_desc = [
        ByteEnumField('type', _TZSPTag.TAG_TYPE_DATA_RATE, _TZSPTag.TAG_TYPES),
        ByteField('len', 1),
        ByteEnumField('data_rate', DATA_RATE_UNKNOWN, DATA_RATES)
    ]


class TZSPTagTimestamp(_TZSPTag):
    """
    MAC receive timestamp
    """
    fields_desc = [
        ByteEnumField('type', _TZSPTag.TAG_TYPE_TIMESTAMP, _TZSPTag.TAG_TYPES),
        ByteField('len', 4),
        IntField('timestamp', 0)
    ]


class TZSPTagContentionFree(_TZSPTag):
    """
    packet received in contention free period
    """
    NO = 0x00
    YES = 0x01

    fields_desc = [
        ByteEnumField('type', _TZSPTag.TAG_TYPE_CONTENTION_FREE, _TZSPTag.TAG_TYPES),  # noqa: E501
        ByteField('len', 1),
        YesNoByteField('contention_free', NO)
    ]


class TZSPTagDecrypted(_TZSPTag):
    """
    packet was decrypted
    """
    YES = 0x00
    NO = 0x01

    fields_desc = [
        ByteEnumField('type', _TZSPTag.TAG_TYPE_DECRYPTED, _TZSPTag.TAG_TYPES),
        ByteField('len', 1),
        YesNoByteField('decrypted', NO, config={'yes': YES, 'no': (NO, 0xff)})
    ]


class TZSPTagError(_TZSPTag):
    """
    frame checksum error
    """
    NO = 0x00
    YES = 0x01

    fields_desc = [
        ByteEnumField('type', _TZSPTag.TAG_TYPE_FCS_ERROR, _TZSPTag.TAG_TYPES),
        ByteField('len', 1),
        YesNoByteField('fcs_error', NO, config={'no': NO, 'yes': YES, 'reserved': (YES + 1, 0xff)})  # noqa: E501
    ]


class TZSPTagRXChannel(_TZSPTag):
    """
    channel the sensor was on while receiving the frame
    """
    fields_desc = [
        ByteEnumField('type', _TZSPTag.TAG_TYPE_RX_CHANNEL, _TZSPTag.TAG_TYPES),  # noqa: E501
        ByteField('len', 1),
        ByteField('rx_channel', 0)
    ]


class TZSPTagPacketCount(_TZSPTag):
    """
    packet counter
    """
    fields_desc = [
        ByteEnumField('type', _TZSPTag.TAG_TYPE_PACKET_COUNT, _TZSPTag.TAG_TYPES),  # noqa: E501
        ByteField('len', 4),
        IntField('packet_count', 0)
    ]


class TZSPTagRXFrameLength(_TZSPTag):
    """
    received packet length
    """
    fields_desc = [
        ByteEnumField('type', _TZSPTag.TAG_TYPE_RX_FRAME_LENGTH, _TZSPTag.TAG_TYPES),  # noqa: E501
        ByteField('len', 2),
        ShortField('rx_frame_length', 0)
    ]


class TZSPTagWlanRadioHdrSerial(_TZSPTag):
    """
    (vendor specific) unique capture device (sensor/AP) identifier
    """
    fields_desc = [
        ByteEnumField('type', _TZSPTag.TAG_TYPE_WLAN_RADIO_HDR_SERIAL, _TZSPTag.TAG_TYPES),  # noqa: E501
        FieldLenField('len', None, length_of='sensor_id', fmt='b'),
        StrLenField('sensor_id', '', length_from=lambda pkt:pkt.len)
    ]


class TZSPTagUnknown(_TZSPTag):
    """
    unknown tag type dummy
    """
    fields_desc = [
        ByteField('type', 0xff),
        FieldLenField('len', None, length_of='data', fmt='b'),
        XStrLenField('data', '', length_from=lambda pkt: pkt.len)
    ]


_TZSP_TAG_CLASSES = {
    _TZSPTag.TAG_TYPE_PADDING: TZSPTagPadding,
    _TZSPTag.TAG_TYPE_END: TZSPTagEnd,
    _TZSPTag.TAG_TYPE_RAW_RSSI: {1: TZSPTagRawRSSIByte, 2: TZSPTagRawRSSIShort},  # noqa: E501
    _TZSPTag.TAG_TYPE_SNR: {1: TZSPTagSNRByte, 2: TZSPTagSNRShort},
    _TZSPTag.TAG_TYPE_DATA_RATE: TZSPTagDataRate,
    _TZSPTag.TAG_TYPE_TIMESTAMP: TZSPTagTimestamp,
    _TZSPTag.TAG_TYPE_CONTENTION_FREE: TZSPTagContentionFree,
    _TZSPTag.TAG_TYPE_DECRYPTED: TZSPTagDecrypted,
    _TZSPTag.TAG_TYPE_FCS_ERROR: TZSPTagError,
    _TZSPTag.TAG_TYPE_RX_CHANNEL: TZSPTagRXChannel,
    _TZSPTag.TAG_TYPE_PACKET_COUNT: TZSPTagPacketCount,
    _TZSPTag.TAG_TYPE_RX_FRAME_LENGTH: TZSPTagRXFrameLength,
    _TZSPTag.TAG_TYPE_WLAN_RADIO_HDR_SERIAL: TZSPTagWlanRadioHdrSerial
}



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information

# scapy.contrib.description = Link Aggregation Control Protocol (LACP)
# scapy.contrib.status = loads

from scapy.packet import Packet, bind_layers
from scapy.fields import ByteField, MACField, ShortField, ByteEnumField, IntField, XStrFixedLenField  # noqa: E501
from scapy.contrib.slowprot import SlowProtocol


class LACP(Packet):
    name = "LACP"
    deprecated_fields = {
        "actor_port_numer": ("actor_port_number", "2.4.4"),
        "partner_port_numer": ("partner_port_number", "2.4.4"),
        "colletctor_reserved": ("collector_reserved", "2.4.4"),
    }
    fields_desc = [
        ByteField("version", 1),
        ByteField("actor_type", 1),
        ByteField("actor_length", 20),
        ShortField("actor_system_priority", 0),
        MACField("actor_system", None),
        ShortField("actor_key", 0),
        ShortField("actor_port_priority", 0),
        ShortField("actor_port_number", 0),
        ByteField("actor_state", 0),
        XStrFixedLenField("actor_reserved", "", 3),
        ByteField("partner_type", 2),
        ByteField("partner_length", 20),
        ShortField("partner_system_priority", 0),
        MACField("partner_system", None),
        ShortField("partner_key", 0),
        ShortField("partner_port_priority", 0),
        ShortField("partner_port_number", 0),
        ByteField("partner_state", 0),
        XStrFixedLenField("partner_reserved", "", 3),
        ByteField("collector_type", 3),
        ByteField("collector_length", 16),
        ShortField("collector_max_delay", 0),
        XStrFixedLenField("collector_reserved", "", 12),
        ByteField("terminator_type", 0),
        ByteField("terminator_length", 0),
        XStrFixedLenField("reserved", "", 50),
    ]


bind_layers(SlowProtocol, LACP, subtype=1)

MARKER_TYPES = {
    'Marker Request': 1,
    'Marker Response': 2,
}


class MarkerProtocol(Packet):
    name = "MarkerProtocol"
    fields_desc = [
        ByteField("version", 1),
        ByteEnumField("marker_type", 1, MARKER_TYPES),
        ByteField("marker_length", 16),
        ShortField("requester_port", 0),
        MACField("requester_system", None),
        IntField("requester_transaction_id", 0),
        XStrFixedLenField("marker_reserved", "", 2),
        ByteField("terminator_type", 0),
        ByteField("terminator_length", 0),
        XStrFixedLenField("reserved", 0, 90),
    ]


bind_layers(SlowProtocol, MarkerProtocol, subtype=2)



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) 2019 Jose Diogo Monteiro <jdlopes@student.dei.uc.pt>
# Updated (C) 2023 Claire Vacherot <clairelex@pm.me>

# scapy.contrib.description = EtherNet/IP
# scapy.contrib.status = loads

"""
EtherNet/IP (Industrial Protocol)

Based on https://github.com/scy-phy/scapy-cip-enip
EtherNet/IP Home: www.odva.org
"""

import struct
from scapy.packet import Packet, bind_layers
from scapy.layers.inet import TCP
from scapy.fields import LEShortField, LEShortEnumField, LEIntEnumField, \
    LEIntField, LELongField, FieldLenField, PacketListField, ByteField, \
    StrLenField, StrFixedLenField, XLEIntField, XLEStrLenField, \
    LEFieldLenField, ShortField, IPField, LongField, XLEShortField

_commandIdList = {
    0x0001: "UnknownCommand",
    0x0004: "ListServices",  # Request Struct Don't Have Command Spec Data
    0x0063: "ListIdentity",  # Request Struct Don't Have Command Spec Data
    0x0064: "ListInterfaces",  # Request Struct Don't Have Command Spec Data
    0x0065: "RegisterSession",  # Request Structure = Reply Structure
    0x0066: "UnregisterSession",  # Don't Have Command Specific Data
    0x006f: "SendRRData",  # Request Structure = Reply Structure
    0x0070: "SendUnitData",  # There is no reply
    0x0072: "IndicateStatus",
    0x0073: "Cancel"
}

_statusList = {
    0: "success",
    1: "invalid_cmd",
    2: "no_resources",
    3: "incorrect_data",
    100: "invalid_session",
    101: "invalid_length",
    105: "unsupported_prot_rev"
}

_typeIdList = {
    0x0000: "Null Address Item",
    0x000c: "CIP Identity",
    0x0086: "CIP Security Information",
    0x0087: "EtherNet/IP Capability",
    0x0088: "EtherNet/IP Usage",
    0x00a1: "Connected Address Item",
    0x00B1: "Connected Data Item",
    0x00B2: "Unconnected Data Item",
    0x0100: "List Services Response",
    0x8000: "Socket Address Info O->T",
    0x8001: "Socket Address Info T->O",
    0x8002: "Sequenced Address Item",
    0x8003: "Unconnected Message over UDP"
}

_deviceTypeList = {
    0x0000: "Generic Device (deprecated)",
    0x0002: "AC Drive",
    0x0003: "Motor Overload",
    0x0004: "Limit Switch",
    0x0005: "Inductive Proximity Switch",
    0x0006: "Photoelectric Sensor",
    0x0007: "General Purpose Discrete I/O",
    0x0009: "Resolver",
    0x000C: "Communications Adapter",
    0x000E: "Programmable Logic Controller",
    0x0010: "Position Controller",
    0x0013: "DC Drive",
    0x0015: "Contactor",
    0x0016: "Motor Starter",
    0x0017: "Soft Start",
    0x0018: "Human-Machine Interface",
    0x001A: "Mass Flow Controller",
    0x001B: "Pneumatic Valve",
    0x001C: "Vacuum Pressure Gauge",
    0x001D: "Process Control Value",
    0x001E: "Residual Gas Analyzer",
    0x001F: "DC Power Generator",
    0x0020: "RF Power Generator",
    0x0021: "Turbomolecular Vacuum Pump",
    0x0022: "Encoder",
    0x0023: "Safety Discrete I/O Device",
    0x0024: "Fluid Flow Controller",
    0x0025: "CIP Motion Drive",
    0x0026: "CompoNet Repeater",
    0x0027: "Mass Flow Controller, Enhanced",
    0x0028: "CIP Modbus Device",
    0x0029: "CIP Modbus Translator",
    0x002A: "Safety Analog I/O Device",
    0x002B: "Generic Device (keyable)",
    0x002C: "Managed Ethernet Switch",
    0x002D: "CIP Motion Safety Drive Device",
    0x002E: "Safety Drive Device",
    0x002F: "CIP Motion Encoder",
    0x0030: "CIP Motion Converter",
    0x0031: "CIP Motion I/O",
    0x0032: "ControlNet Physical Layer Component",
    0x0033: "Circuit Breaker",
    0x0034: "HART Device",
    0x0035: "CIP-HART Translator",
    0x00C8: "Embedded Component",
}

_interfaceList = {
    0x00: "CIP"
}


class ItemData(Packet):
    """Common Packet Format"""
    name = "Item Data"
    fields_desc = [
        LEShortEnumField("typeId", 0, _typeIdList),
        LEShortField("length", 0),
        XLEStrLenField("data", "", length_from=lambda pkt: pkt.length),
    ]

    def extract_padding(self, s):
        return '', s


# Unknown command (0x0001)


class ENIPUnknownCommand(Packet):
    """Unknown Command reply"""
    name = "ENIPUnknownCommand"
    pass


# List services (0x0004)


class ENIPListServicesItem(Packet):
    """List Services Item Field"""
    name = "ENIPListServicesItem"
    fields_desc = [
        LEShortEnumField("itemTypeCode", 0, _typeIdList),
        LEFieldLenField("itemLength", 0),
        LEShortField("protocolVersion", 0),
        XLEShortField("flag", 0),  # TODO: detail with BitFields
        StrFixedLenField("serviceName", None, 16),
    ]


class ENIPListServices(Packet):
    """List Services Command Field"""
    name = "ENIPListServices"
    fields_desc = [
        FieldLenField("itemCount", 0, count_of="items"),
        PacketListField("items", None, ENIPListServicesItem),
    ]


# List identity (0x0063)


class ENIPListIdentityItem(Packet):
    """List Identity Item Fields"""
    name = "ENIPListIdentityReplyItem"
    fields_desc = [
        LEShortEnumField("itemTypeCode", 0, _typeIdList),
        LEFieldLenField("itemLength", 0),
        LEShortField("protocolVersion", 0),
        # Socket address
        ShortField("sinFamily", 0),
        ShortField("sinPort", 0),
        IPField("sinAddress", None),
        LongField("sinZero", 0),
        # End socket address
        LEShortField("vendorId", 0),  # Vendor list could be added (long list)
        LEShortEnumField("deviceType", 0, _deviceTypeList),
        LEShortField("productCode", 0),
        ByteField("revisionMajor", 0),
        ByteField("revisionMinor", 0),
        LEShortField("status", 0),
        XLEIntField("serialNumber", 0),
        ByteField("productNameLength", 0),
        StrLenField("productName", None,
                    length_from=lambda pkt: pkt.productNameLength),
        ByteField("state", 0)
    ]


class ENIPListIdentity(Packet):
    """List identity request and response"""
    name = "ENIPListIdentity"
    fields_desc = [
        FieldLenField("itemCount", 0, count_of="items"),
        PacketListField("items", None, ENIPListIdentityItem)
    ]


# List Interfaces (0x0064)


class ENIPListInterfacesItem(Packet):
    """List Interfaces Item Fields"""
    name = "ENIPListInterfacesItem"
    fields_desc = [
        LEShortEnumField("itemTypeCode", 0, _typeIdList),
        FieldLenField("itemLength", 0, length_of="itemData"),
        # TODO: Could be detailed
        StrLenField("itemData", "", length_from=lambda pkt: pkt.itemLength),
    ]


class ENIPListInterfaces(Packet):
    """List Interfaces Command Field"""
    name = "ENIPListInterfaces"
    fields_desc = [
        FieldLenField("itemCount", 0, count_of="items"),
        PacketListField("items", None, ENIPListInterfacesItem),
    ]


# Register Session (0x0065)


class ENIPRegisterSession(Packet):
    """Register Session Command Field"""
    name = "ENIPRegisterSession"
    fields_desc = [
        LEShortField("protocolVersion", 1),
        LEShortField("options", 0)
    ]


# Unregister Session (0x0066) -- Requires further testing


class ENIPUnregisterSession(Packet):
    """Unregister Session Command Field"""
    name = "ENIPUnregisterSession"
    pass


# Send RR Data (0x006f)


class ENIPSendRRData(Packet):
    """Send RR Data Command Field"""
    name = "ENIPSendRRData"
    fields_desc = [
        LEIntEnumField("interface", 0, _interfaceList),
        LEShortField("timeout", 0xff),
        LEFieldLenField("itemCount", 0, count_of="items"),
        PacketListField("items", None, ItemData)
        # TODO: Send RR Data is usually followed by a CIP packet
    ]


# Send Unit Data (0x006f)


class ENIPSendUnitData(Packet):
    """Send Unit Data Command Field"""
    name = "ENIPSendUnitData"
    fields_desc = [
        LEIntEnumField("interface", 0, _interfaceList),
        LEShortField("timeout", 0xff),
        LEFieldLenField("itemCount", 0, count_of="items"),
        PacketListField("items", None, ItemData)
    ]


# Main Ethernet/IP packet structure with header


class ENIPTCP(Packet):
    """Ethernet/IP packet over TCP"""
    name = "ENIPTCP"
    fields_desc = [
        LEShortEnumField("commandId", None, _commandIdList),
        LEShortField("length", 0),
        XLEIntField("session", 0),
        LEIntEnumField("status", None, _statusList),
        LELongField("senderContext", 0),
        LEIntField("options", 0),
    ]

    def post_build(self, pkt, pay):
        if self.length is None and pay:
            pkt = pkt[:2] + struct.pack("<H", len(pay)) + pkt[4:]
        return pkt + pay


bind_layers(TCP, ENIPTCP, dport=44818)
bind_layers(TCP, ENIPTCP, sport=44818)

bind_layers(ENIPTCP, ENIPUnknownCommand, commandId=0x0001)
bind_layers(ENIPTCP, ENIPListServices, commandId=0x0004)
bind_layers(ENIPTCP, ENIPListIdentity, commandId=0x0063)
bind_layers(ENIPTCP, ENIPListInterfaces, commandId=0x0064)
bind_layers(ENIPTCP, ENIPRegisterSession, commandId=0x0065)
bind_layers(ENIPTCP, ENIPUnregisterSession, commandId=0x0066)
bind_layers(ENIPTCP, ENIPSendRRData, commandId=0x006f)
bind_layers(ENIPTCP, ENIPSendUnitData, commandId=0x0070)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>

"""
Routing and handling of network interfaces.
"""


from scapy.compat import plain_str
from scapy.config import conf
from scapy.error import Scapy_Exception, warning
from scapy.interfaces import resolve_iface
from scapy.utils import atol, ltoa, itom, pretty_list

from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)


##############################
#  Routing/Interfaces stuff  #
##############################

class Route:
    def __init__(self):
        # type: () -> None
        self.routes = []  # type: List[Tuple[int, int, str, str, str, int]]
        self.invalidate_cache()
        if conf.route_autoload:
            self.resync()

    def invalidate_cache(self):
        # type: () -> None
        self.cache = {}  # type: Dict[Tuple[str, Optional[str]], Tuple[str, str, str]]

    def resync(self):
        # type: () -> None
        from scapy.arch import read_routes
        self.invalidate_cache()
        self.routes = read_routes()

    def __repr__(self):
        # type: () -> str
        rtlst = []  # type: List[Tuple[Union[str, List[str]], ...]]
        for net, msk, gw, iface, addr, metric in self.routes:
            if_repr = resolve_iface(iface).description
            rtlst.append((ltoa(net),
                          ltoa(msk),
                          gw,
                          if_repr,
                          addr,
                          str(metric)))

        return pretty_list(rtlst,
                           [("Network", "Netmask", "Gateway", "Iface", "Output IP", "Metric")])  # noqa: E501

    def make_route(self,
                   host=None,  # type: Optional[str]
                   net=None,  # type: Optional[str]
                   gw=None,  # type: Optional[str]
                   dev=None,  # type: Optional[str]
                   metric=1,  # type: int
                   ):
        # type: (...) -> Tuple[int, int, str, str, str, int]
        if host is not None:
            thenet, msk = host, 32
        elif net is not None:
            thenet, msk_b = net.split("/")
            msk = int(msk_b)
        else:
            raise Scapy_Exception("make_route: Incorrect parameters. You should specify a host or a net")  # noqa: E501
        if gw is None:
            gw = "0.0.0.0"
        if dev is None:
            if gw:
                nhop = gw
            else:
                nhop = thenet
            dev, ifaddr, _ = self.route(nhop)
        else:
            ifaddr = "0.0.0.0"  # acts as a 'via' in `ip addr add`
        return (atol(thenet), itom(msk), gw, dev, ifaddr, metric)

    def add(self, *args, **kargs):
        # type: (*Any, **Any) -> None
        """Add a route to Scapy's IPv4 routing table.
        add(host|net, gw|dev)

        :param host: single IP to consider (/32)
        :param net: range to consider
        :param gw: gateway
        :param dev: force the interface to use
        :param metric: route metric

        Examples:

        - `ip route add 192.168.1.0/24 via 192.168.0.254`::
            >>> conf.route.add(net="192.168.1.0/24", gw="192.168.0.254")

        - `ip route add 192.168.1.0/24 dev eth0`::
            >>> conf.route.add(net="192.168.1.0/24", dev="eth0")

        - `ip route add 192.168.1.0/24 via 192.168.0.254 metric 1`::
            >>> conf.route.add(net="192.168.1.0/24", gw="192.168.0.254", metric=1)
        """
        self.invalidate_cache()
        self.routes.append(self.make_route(*args, **kargs))

    def delt(self, *args, **kargs):
        # type: (*Any, **Any) -> None
        """Remove a route from Scapy's IPv4 routing table.
        delt(host|net, gw|dev)

        Same syntax as add()
        """
        self.invalidate_cache()
        route = self.make_route(*args, **kargs)
        try:
            i = self.routes.index(route)
            del self.routes[i]
        except ValueError:
            raise ValueError("No matching route found!")

    def ifchange(self, iff, addr):
        # type: (str, str) -> None
        self.invalidate_cache()
        the_addr, the_msk_b = (addr.split("/") + ["32"])[:2]
        the_msk = itom(int(the_msk_b))
        the_rawaddr = atol(the_addr)
        the_net = the_rawaddr & the_msk

        for i, route in enumerate(self.routes):
            net, msk, gw, iface, addr, metric = route
            if iff != iface:
                continue
            if gw == '0.0.0.0':
                self.routes[i] = (the_net, the_msk, gw, iface, the_addr, metric)  # noqa: E501
            else:
                self.routes[i] = (net, msk, gw, iface, the_addr, metric)
        conf.netcache.flush()

    def ifdel(self, iff):
        # type: (str) -> None
        self.invalidate_cache()
        new_routes = []
        for rt in self.routes:
            if iff == rt[3]:
                continue
            new_routes.append(rt)
        self.routes = new_routes

    def ifadd(self, iff, addr):
        # type: (str, str) -> None
        self.invalidate_cache()
        the_addr, the_msk_b = (addr.split("/") + ["32"])[:2]
        the_msk = itom(int(the_msk_b))
        the_rawaddr = atol(the_addr)
        the_net = the_rawaddr & the_msk
        self.routes.append((the_net, the_msk, '0.0.0.0', iff, the_addr, 1))

    def route(self, dst=None, dev=None, verbose=conf.verb, _internal=False):
        # type: (Optional[str], Optional[str], int, bool) -> Tuple[str, str, str]
        """Returns the IPv4 routes to a host.

        :param dst: the IPv4 of the destination host
        :param dev: (optional) filtering is performed to limit search to route
                    associated to that interface.

        :returns: tuple (iface, output_ip, gateway_ip) where
            - ``iface``: the interface used to connect to the host
            - ``output_ip``: the outgoing IP that will be used
            - ``gateway_ip``: the gateway IP that will be used
        """
        dst = dst or "0.0.0.0"  # Enable route(None) to return default route
        if isinstance(dst, bytes):
            try:
                dst = plain_str(dst)
            except UnicodeDecodeError:
                raise TypeError("Unknown IP address input (bytes)")
        if (dst, dev) in self.cache:
            return self.cache[(dst, dev)]
        # Transform "192.168.*.1-5" to one IP of the set
        _dst = dst.split("/")[0].replace("*", "0")
        while True:
            idx = _dst.find("-")
            if idx < 0:
                break
            m = (_dst[idx:] + ".").find(".")
            _dst = _dst[:idx] + _dst[idx + m:]

        atol_dst = atol(_dst)
        paths = []
        for d, m, gw, i, a, me in self.routes:
            if not a:  # some interfaces may not currently be connected
                continue
            if dev is not None and i != dev:
                continue
            aa = atol(a)
            if aa == atol_dst and aa != 0:
                paths.append(
                    (0xffffffff, 1, (conf.loopback_name, a, "0.0.0.0"))  # noqa: E501
                )
            if (atol_dst & m) == (d & m):
                paths.append((m, me, (i, a, gw)))

        if not paths:
            if verbose:
                warning("No route found for IPv4 destination %s "
                        "(no default route?)", dst)
            return (dev or conf.loopback_name, "0.0.0.0", "0.0.0.0")
        # Choose the more specific route
        # Sort by greatest netmask and use metrics as a tie-breaker
        paths.sort(key=lambda x: (-x[0], x[1]))
        # Return interface
        ret = paths[0][2]
        # Check if source is 0.0.0.0. This is a 'via' route with no src.
        if ret[1] == "0.0.0.0" and not _internal:
            # Then get the source from route(gw)
            ret = (ret[0], self.route(ret[2], _internal=True)[1], ret[2])
        self.cache[(dst, dev)] = ret
        return ret

    def get_if_bcast(self, iff):
        # type: (str) -> List[str]
        """
        Return the list of broadcast addresses of an interface.
        """
        bcast_list = []
        for net, msk, _, iface, _, _ in self.routes:
            if net == 0:
                continue    # Ignore default route "0.0.0.0"
            elif msk == 0xffffffff:
                continue    # Ignore host-specific routes
            if iff != iface:
                continue
            bcast = net | (~msk & 0xffffffff)
            bcast_list.append(ltoa(bcast))
        if not bcast_list:
            warning("No broadcast address found for iface %s\n", iff)
        return bcast_list


conf.route = Route()

# Update conf.iface
conf.ifaces.load_confiface()



#! /usr/bin/env python

# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Nils Weiss <nils@we155.de>

# scapy.contrib.description = Diagnostic over IP (DoIP) / ISO 13400
# scapy.contrib.status = loads

import socket
import ssl
import struct
import time
from typing import (
    Any,
    Union,
    Tuple,
    Optional,
    Dict,
    Type,
)

from scapy.contrib.automotive import log_automotive
from scapy.contrib.automotive.uds import UDS
from scapy.data import MTU
from scapy.fields import (
    ByteEnumField,
    ConditionalField,
    IntField,
    MayEnd,
    StrFixedLenField,
    XByteEnumField,
    XByteField,
    XIntField,
    XShortEnumField,
    XShortField,
    XStrField,
)
from scapy.layers.inet import TCP, UDP
from scapy.packet import Packet, bind_layers, bind_bottom_up
from scapy.supersocket import SSLStreamSocket


# ISO 13400-2 sect 9.2


class DoIP(Packet):
    """
    Implementation of the DoIP (ISO 13400) protocol. DoIP packets can be sent
    via UDP and TCP. Depending on the payload type, the correct connection
    need to be chosen:

    +--------------+--------------------------------------------------------------+-----------------+
    | Payload Type | Payload Type Name                                            | Connection Kind |
    +--------------+--------------------------------------------------------------+-----------------+
    | 0x0000       | Generic DoIP header negative acknowledge                     | UDP / TCP       |
    +--------------+--------------------------------------------------------------+-----------------+
    | 0x0001       | Vehicle Identification request message                       | UDP             |
    +--------------+--------------------------------------------------------------+-----------------+
    | 0x0002       | Vehicle identification request message with EID              | UDP             |
    +--------------+--------------------------------------------------------------+-----------------+
    | 0x0003       | Vehicle identification request message with VIN              | UDP             |
    +--------------+--------------------------------------------------------------+-----------------+
    | 0x0004       | Vehicle announcement message/vehicle identification response | UDP             |
    +--------------+--------------------------------------------------------------+-----------------+
    | 0x0005       | Routing activation request                                   | TCP             |
    +--------------+--------------------------------------------------------------+-----------------+
    | 0x0006       | Routing activation response                                  | TCP             |
    +--------------+--------------------------------------------------------------+-----------------+
    | 0x0007       | Alive Check request                                          | TCP             |
    +--------------+--------------------------------------------------------------+-----------------+
    | 0x0008       | Alive Check response                                         | TCP             |
    +--------------+--------------------------------------------------------------+-----------------+
    | 0x4001       | IP entity status request                                     | UDP             |
    +--------------+--------------------------------------------------------------+-----------------+
    | 0x4002       | DoIP entity status response                                  | UDP             |
    +--------------+--------------------------------------------------------------+-----------------+
    | 0x4003       | Diagnostic power mode information request                    | UDP             |
    +--------------+--------------------------------------------------------------+-----------------+
    | 0x4004       | Diagnostic power mode information response                   | UDP             |
    +--------------+--------------------------------------------------------------+-----------------+
    | 0x8001       | Diagnostic message                                           | TCP             |
    +--------------+--------------------------------------------------------------+-----------------+
    | 0x8002       | Diagnostic message positive acknowledgement                  | TCP             |
    +--------------+--------------------------------------------------------------+-----------------+
    | 0x8003       | Diagnostic message negative acknowledgement                  | TCP             |
    +--------------+--------------------------------------------------------------+-----------------+

    Example with UDP:
        >>> socket = L3RawSocket(iface="eth0")
        >>> resp = socket.sr1(IP(dst="169.254.117.238")/UDP(dport=13400)/DoIP(payload_type=1))

    Example with TCP:
        >>> socket = DoIPSocket("169.254.117.238")
        >>> pkt = DoIP(payload_type=0x8001, source_address=0xe80, target_address=0x1000) / UDS() / UDS_RDBI(identifiers=[0x1000])
        >>> resp = socket.sr1(pkt, timeout=1)

    Example with UDS:
        >>> socket = UDS_DoIPSocket("169.254.117.238")
        >>> pkt = UDS() / UDS_RDBI(identifiers=[0x1000])
        >>> resp = socket.sr1(pkt, timeout=1)
    """  # noqa: E501
    payload_types = {
        0x0000: "Generic DoIP header NACK",
        0x0001: "Vehicle identification request",
        0x0002: "Vehicle identification request with EID",
        0x0003: "Vehicle identification request with VIN",
        0x0004: "Vehicle announcement message/vehicle identification response message",  # noqa: E501
        0x0005: "Routing activation request",
        0x0006: "Routing activation response",
        0x0007: "Alive check request",
        0x0008: "Alive check response",
        0x4001: "DoIP entity status request",
        0x4002: "DoIP entity status response",
        0x4003: "Diagnostic power mode information request",
        0x4004: "Diagnostic power mode information response",
        0x8001: "Diagnostic message",
        0x8002: "Diagnostic message ACK",
        0x8003: "Diagnostic message NACK"}
    name = 'DoIP'
    fields_desc = [
        XByteEnumField("protocol_version", 0x02, {
            0x01: "ISO13400_2010", 0x02: "ISO13400_2012",
            0x03: "ISO13400_2019", 0x04: "ISO13400_2019_AMD1"}),
        XByteEnumField("inverse_version", 0xFD, {
            0xFE: "ISO13400_2010", 0xFD: "ISO13400_2012",
            0xFC: "ISO13400_2019", 0xFB: "ISO13400_2019_AMD1"}),
        XShortEnumField("payload_type", 0, payload_types),
        IntField("payload_length", None),
        ConditionalField(ByteEnumField("nack", 0, {
            0: "Incorrect pattern format", 1: "Unknown payload type",
            2: "Message too large", 3: "Out of memory",
            4: "Invalid payload length"
        }), lambda p: p.payload_type in [0x0]),
        ConditionalField(StrFixedLenField("vin", b"", 17),
                         lambda p: p.payload_type in [3, 4]),
        ConditionalField(XShortField("logical_address", 0),
                         lambda p: p.payload_type in [4]),
        ConditionalField(StrFixedLenField("eid", b"", 6),
                         lambda p: p.payload_type in [2, 4]),
        ConditionalField(StrFixedLenField("gid", b"", 6),
                         lambda p: p.payload_type in [4]),
        ConditionalField(MayEnd(XByteEnumField("further_action", 0, {
            0x00: "No further action required",
            0x01: "Reserved by ISO 13400", 0x02: "Reserved by ISO 13400",
            0x03: "Reserved by ISO 13400", 0x04: "Reserved by ISO 13400",
            0x05: "Reserved by ISO 13400", 0x06: "Reserved by ISO 13400",
            0x07: "Reserved by ISO 13400", 0x08: "Reserved by ISO 13400",
            0x09: "Reserved by ISO 13400", 0x0a: "Reserved by ISO 13400",
            0x0b: "Reserved by ISO 13400", 0x0c: "Reserved by ISO 13400",
            0x0d: "Reserved by ISO 13400", 0x0e: "Reserved by ISO 13400",
            0x0f: "Reserved by ISO 13400",
            0x10: "Routing activation required to initiate central security",
        })), lambda p: p.payload_type in [4]),
        # VIN/GID sync. status is marked as optional, so the packet MayEnd
        # on further_action
        ConditionalField(XByteEnumField("vin_gid_status", 0, {
            0x00: "VIN and/or GID are synchronized",
            0x01: "Reserved by ISO 13400", 0x02: "Reserved by ISO 13400",
            0x03: "Reserved by ISO 13400", 0x04: "Reserved by ISO 13400",
            0x05: "Reserved by ISO 13400", 0x06: "Reserved by ISO 13400",
            0x07: "Reserved by ISO 13400", 0x08: "Reserved by ISO 13400",
            0x09: "Reserved by ISO 13400", 0x0a: "Reserved by ISO 13400",
            0x0b: "Reserved by ISO 13400", 0x0c: "Reserved by ISO 13400",
            0x0d: "Reserved by ISO 13400", 0x0e: "Reserved by ISO 13400",
            0x0f: "Reserved by ISO 13400",
            0x10: "Incomplete: VIN and GID are NOT synchronized"
        }), lambda p: p.payload_type in [4]),
        ConditionalField(XShortField("source_address", 0),
                         lambda p: p.payload_type in [5, 8, 0x8001, 0x8002, 0x8003]),  # noqa: E501
        ConditionalField(XByteEnumField("activation_type", 0, {
            0: "Default", 1: "WWH-OBD", 0xe0: "Central security",
            0x16: "Default", 0x116: "Diagnostic", 0xe016: "Central security"
        }), lambda p: p.payload_type in [5]),
        ConditionalField(XShortField("logical_address_tester", 0),
                         lambda p: p.payload_type in [6]),
        ConditionalField(XShortField("logical_address_doip_entity", 0),
                         lambda p: p.payload_type in [6]),
        ConditionalField(XByteEnumField("routing_activation_response", 0, {
            0x00: "Routing activation denied due to unknown source address.",
            0x01: "Routing activation denied because all concurrently supported TCP_DATA sockets are registered and active.",  # noqa: E501
            0x02: "Routing activation denied because an SA different from the table connection entry was received on the already activated TCP_DATA socket.",  # noqa: E501
            0x03: "Routing activation denied because the SA is already registered and active on a different TCP_DATA socket.",  # noqa: E501
            0x04: "Routing activation denied due to missing authentication.",
            0x05: "Routing activation denied due to rejected confirmation.",
            0x06: "Routing activation denied due to unsupported routing activation type.",  # noqa: E501
            0x07: "Routing activation denied because the specified activation type requires a secure TLS TCP_DATA socket.",  # noqa: E501
            0x08: "Reserved by ISO 13400.",
            0x09: "Reserved by ISO 13400.", 0x0a: "Reserved by ISO 13400.",
            0x0b: "Reserved by ISO 13400.", 0x0c: "Reserved by ISO 13400.",
            0x0d: "Reserved by ISO 13400.", 0x0e: "Reserved by ISO 13400.",
            0x0f: "Reserved by ISO 13400.",
            0x10: "Routing successfully activated.",
            0x11: "Routing will be activated; confirmation required."
        }), lambda p: p.payload_type in [6]),
        ConditionalField(XIntField("reserved_iso", 0),
                         lambda p: p.payload_type in [5, 6]),
        ConditionalField(XStrField("reserved_oem", b""),
                         lambda p: p.payload_type in [5, 6]),
        ConditionalField(XByteEnumField("diagnostic_power_mode", 0, {
            0: "not ready", 1: "ready", 2: "not supported"
        }), lambda p: p.payload_type in [0x4004]),
        ConditionalField(ByteEnumField("node_type", 0, {
            0: "DoIP gateway", 1: "DoIP node"
        }), lambda p: p.payload_type in [0x4002]),
        ConditionalField(XByteField("max_open_sockets", 1),
                         lambda p: p.payload_type in [0x4002]),
        ConditionalField(XByteField("cur_open_sockets", 0),
                         lambda p: p.payload_type in [0x4002]),
        ConditionalField(IntField("max_data_size", 0),
                         lambda p: p.payload_type in [0x4002]),
        ConditionalField(XShortField("target_address", 0),
                         lambda p: p.payload_type in [0x8001, 0x8002, 0x8003]),  # noqa: E501
        ConditionalField(XByteEnumField("ack_code", 0, {0: "ACK"}),
                         lambda p: p.payload_type in [0x8002]),
        ConditionalField(ByteEnumField("nack_code", 0, {
            0x00: "Reserved by ISO 13400", 0x01: "Reserved by ISO 13400",
            0x02: "Invalid source address", 0x03: "Unknown target address",
            0x04: "Diagnostic message too large", 0x05: "Out of memory",
            0x06: "Target unreachable", 0x07: "Unknown network",
            0x08: "Transport protocol error"
        }), lambda p: p.payload_type in [0x8003]),
        ConditionalField(XStrField("previous_msg", b""),
                         lambda p: p.payload_type in [0x8002, 0x8003])
    ]

    def answers(self, other):
        # type: (Packet) -> int
        """DEV: true if self is an answer from other"""
        if isinstance(other, type(self)):
            if self.payload_type == 0:
                return 1

            matches = [(4, 1), (4, 2), (4, 3), (6, 5), (8, 7),
                       (0x4002, 0x4001), (0x4004, 0x4003),
                       (0x8001, 0x8001), (0x8003, 0x8001)]
            if (self.payload_type, other.payload_type) in matches:
                if self.payload_type == 0x8001:
                    return self.payload.answers(other.payload)
                return 1
        return 0

    def hashret(self):
        # type: () -> bytes
        payload_type_mapping = {
            0x0000: b"\x01",
            0x0001: b"\x01",
            0x0002: b"\x01",
            0x0003: b"\x01",
            0x0004: b"\x01",
            0x0005: b"\x02",
            0x0006: b"\x02",
            0x0007: b"\x03",
            0x0008: b"\x03",
            0x4001: b"\x04",
            0x4002: b"\x04",
            0x4003: b"\x05",
            0x4004: b"\x05",
            0x8001: b"\x06",
            0x8002: b"\x06",
            0x8003: b"\x06",
        }

        return payload_type_mapping.get(self.payload_type, b"\xff")

    def post_build(self, pkt, pay):
        # type: (bytes, bytes) -> bytes
        """
        This will set the Field 'payload_length' to the correct value.
        """
        if self.payload_length is None:
            pkt = pkt[:4] + struct.pack(
                "!I", len(pay) + len(pkt) - 8) + pkt[8:]
        return pkt + pay

    def extract_padding(self, s):
        # type: (bytes) -> Tuple[bytes, Optional[bytes]]
        if self.payload_type == 0x8001:
            return s[:self.payload_length - 4], s[self.payload_length - 4:]
        else:
            return b"", s

    @classmethod
    def tcp_reassemble(cls, data, metadata, session):
        # type: (bytes, Dict[str, Any], Dict[str, Any]) -> Optional[Packet]
        length = struct.unpack("!I", data[4:8])[0] + 8
        if len(data) >= length:
            return DoIP(data)
        return None


bind_bottom_up(UDP, DoIP, sport=13400)
bind_bottom_up(UDP, DoIP, dport=13400)
bind_layers(UDP, DoIP, sport=13400, dport=13400)

bind_layers(TCP, DoIP, sport=13400)
bind_layers(TCP, DoIP, dport=13400)

bind_layers(DoIP, UDS, payload_type=0x8001)


class DoIPSSLStreamSocket(SSLStreamSocket):
    """Custom SSLStreamSocket for DoIP communication.
    """

    def __init__(self, sock, basecls=None):
        # type: (socket.socket, Optional[Type[Packet]]) -> None
        super(DoIPSSLStreamSocket, self).__init__(sock, basecls or DoIP)
        self.buffer = b""

    def recv(self, x=MTU, **kwargs):
        # type: (Optional[int], **Any) -> Optional[Packet]
        if len(self.buffer) < 8:
            self.buffer += self.ins.recv(8)
        if len(self.buffer) < 8:
            return None
        len_data = self.buffer[:8]

        len_int = struct.unpack(">I", len_data[4:8])[0]
        len_int += 8

        self.buffer += self.ins.recv(len_int - len(self.buffer))
        if len(self.buffer) < len_int:
            return None
        pktbuf = self.buffer[:len_int]
        self.buffer = self.buffer[len_int:]

        pkt = self.basecls(pktbuf, **kwargs)  # type: Packet
        return pkt


class DoIPSocket(DoIPSSLStreamSocket):
    """Socket for DoIP communication. This sockets automatically
    sends a routing activation request as soon as a TCP or TLS connection is
    established.

    :param ip: IP address of destination
    :param port: destination port, usually 13400
    :param tls_port: destination port for TLS connection, usually 3496
    :param activate_routing: If true, routing activation request is
                             automatically sent
    :param source_address: DoIP source address
    :param target_address: DoIP target address, this is automatically
                           determined if routing activation request is sent
    :param activation_type: This allows to set a different activation type for
                            the routing activation request
    :param reserved_oem: Optional parameter to set value for reserved_oem field
                         of routing activation request
    :param force_tls: Skip establishing of a TCP connection and directly try to
                      connect via SSL/TLS
    :param context: Optional ssl.SSLContext object for initialization of ssl socket
                    connections.
    :param doip_version: DoIP protocol version to use, default is 2 (ISO 13400-2012)
    :param enforce_doip_version: If true, the protocol_version field in each DoIP
                                 packet to be sent, is always set to the value of
                                 doip_version.

    Example:
        >>> socket = DoIPSocket("169.254.0.131")
        >>> pkt = DoIP(payload_type=0x8001, source_address=0xe80, target_address=0x1000) / UDS() / UDS_RDBI(identifiers=[0x1000])
        >>> resp = socket.sr1(pkt, timeout=1)
    """  # noqa: E501

    def __init__(self,
                 ip='127.0.0.1',  # type: str
                 port=13400,  # type: int
                 tls_port=3496,  # type: int
                 activate_routing=True,  # type: bool
                 source_address=0xe80,  # type: int
                 target_address=0,  # type: int
                 activation_type=0,  # type: int
                 reserved_oem=b"",  # type: bytes
                 force_tls=False,  # type: bool
                 context=None,  # type: Optional[ssl.SSLContext]
                 doip_version=2,  # type: int
                 enforce_doip_version=False,  # type: bool
                 ):  # type: (...) -> None
        self.ip = ip
        self.port = port
        self.tls_port = tls_port
        self.activate_routing = activate_routing
        self.source_address = source_address
        self.target_address = target_address
        self.activation_type = activation_type
        self.reserved_oem = reserved_oem
        self.force_tls = force_tls
        self.context = context
        self.doip_version = doip_version
        self.enforce_doip_version = enforce_doip_version
        try:
            self._init_socket()
        except Exception:
            self.close()
            raise

    def _init_socket(self):
        # type: () -> None
        connected = False
        addrinfo = socket.getaddrinfo(self.ip, self.port, proto=socket.IPPROTO_TCP)
        sock_family = addrinfo[0][0]

        s = socket.socket(sock_family, socket.SOCK_STREAM)
        s.settimeout(5)
        s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        if not self.force_tls:
            s.connect(addrinfo[0][-1])
            connected = True
            DoIPSSLStreamSocket.__init__(self, s)

            if not self.activate_routing:
                return

            activation_return = self._activate_routing()
        else:
            # Let's overwrite activation_return to force TLS Connection
            activation_return = 0x07

        if activation_return == 0x10:
            # Routing successfully activated.
            return
        elif activation_return == 0x07:
            # Routing activation denied because the specified activation
            # type requires a secure TLS TCP_DATA socket.
            if self.context is None:
                raise ValueError("SSLContext 'context' can not be None")
            if connected:
                s.close()
                s = socket.socket(sock_family, socket.SOCK_STREAM)
                s.settimeout(5)
                s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

            ss = self.context.wrap_socket(s)
            addrinfo = socket.getaddrinfo(
                self.ip, self.tls_port, proto=socket.IPPROTO_TCP)
            ss.connect(addrinfo[0][-1])
            DoIPSSLStreamSocket.__init__(self, ss)

            if not self.activate_routing:
                return

            activation_return = self._activate_routing()
            if activation_return == 0x10:
                # Routing successfully activated.
                return
            else:
                raise Exception(
                    "DoIPSocket activate_routing failed with "
                    "routing_activation_response 0x%x" % activation_return)

        elif activation_return == -1:
            raise Exception("DoIPSocket._activate_routing failed")
        else:
            raise Exception(
                "DoIPSocket activate_routing failed with "
                "routing_activation_response 0x%x!" % activation_return)

    def _activate_routing(self):  # type: (...) -> int
        resp = self.sr1(
            DoIP(payload_type=0x5, activation_type=self.activation_type,
                 source_address=self.source_address, reserved_oem=self.reserved_oem),
            verbose=False, timeout=1)
        if resp and resp.payload_type == 0x6 and \
                resp.routing_activation_response == 0x10:
            self.target_address = (
                self.target_address or resp.logical_address_doip_entity)
            log_automotive.info(
                "Routing activation successful! Target address set to: 0x%x",
                self.target_address)
        else:
            log_automotive.error(
                "Routing activation failed! Response: %s", repr(resp))

        if resp and resp.payload_type == 0x6:
            return resp.routing_activation_response
        else:
            return -1

    def send(self, x):  # type: (Packet) -> int
        if self.enforce_doip_version and isinstance(x, DoIP):
            x[DoIP].protocol_version = self.doip_version
            x[DoIP].inverse_version = 0xFF - self.doip_version
        return super().send(x)


class UDS_DoIPSocket(DoIPSocket):
    """
    Application-Layer socket for DoIP endpoints. This socket takes care about
    the encapsulation of UDS packets into DoIP packets.

    Example:
        >>> socket = UDS_DoIPSocket("169.254.117.238")
        >>> pkt = UDS() / UDS_RDBI(identifiers=[0x1000])
        >>> resp = socket.sr1(pkt, timeout=1)
    """

    def send(self, x):
        # type: (Union[Packet, bytes]) -> int
        if isinstance(x, UDS):
            pkt = DoIP(payload_type=0x8001,
                       source_address=self.source_address,
                       target_address=self.target_address
                       ) / x
        else:
            pkt = x

        try:
            x.sent_time = time.time()  # type: ignore
        except AttributeError:
            pass

        return super().send(pkt)

    def recv(self, x=MTU, **kwargs):
        # type: (Optional[int], **Any) -> Optional[Packet]
        pkt = super().recv(x, **kwargs)
        if pkt and pkt.payload_type == 0x8001:
            return pkt.payload
        else:
            return pkt

    pass
